from ltr.actors import BaseActor
import torch
import numpy as np

def iou_cal(box1, box2, wh=False):
    """
    compute the iou of two boxes.
    Args:
        box1, box2: [xmin, ymin, xmax, ymax] (wh=False) or [xcenter, ycenter, w, h] (wh=True)
        wh: the format of coordinate.
    Return:
        iou: iou of box1 and box2
    """
    N = box1.shape[0]
    iou_total = []
    for i in range(N):
        if wh == False:
            xmin1, ymin1, xmax1, ymax1 = box1[i, 0].clone().detach().item(), box1[i, 1].clone().detach().item(), box1[i, 2].clone().detach().item(), box1[i, 3].clone().detach().item()
            xmin2, ymin2, xmax2, ymax2 = box2[i, 0].clone().detach().item(), box2[i, 1].clone().detach().item(), box2[i, 2].clone().detach().item(), box2[i, 3].clone().detach().item()
        else:
            xmin1, ymin1 = int(box1[i, 0].clone().detach().item() - box1[i, 2].clone().detach().item() / 2.0), int(box1[i, 1].clone().detach().item() - box1[i, 3].clone().detach().item() / 2.0)
            xmax1, ymax1 = int(box1[i, 0].clone().detach().item() + box1[i, 2].clone().detach().item() / 2.0), int(box1[i, 1].clone().detach().item() + box1[i, 3].clone().detach().item() / 2.0)
            xmin2, ymin2 = int(box2[i, 0].clone().detach().item() - box2[i, 2].clone().detach().item() / 2.0), int(box2[i, 1].clone().detach().item() - box2[i, 3].clone().detach().item() / 2.0)
            xmax2, ymax2 = int(box2[i, 0].clone().detach().item() + box2[i, 2].clone().detach().item() / 2.0), int(box2[i, 1].clone().detach().item() + box2[i, 3].clone().detach().item() / 2.0)

        xx1 = np.max([xmin1, xmin2])
        yy1 = np.max([ymin1, ymin2])
        xx2 = np.min([xmax1, xmax2])
        yy2 = np.min([ymax1, ymax2])

        area1 = (xmax1 - xmin1) * (ymax1 - ymin1)
        area2 = (xmax2 - xmin2) * (ymax2 - ymin2)

        inter_area = (np.max([0, xx2 - xx1])) * (np.max([0, yy2 - yy1]))
        iou = inter_area / (area1 + area2 - inter_area + 1e-6)

        iou_total.append(iou)

    iou_mean = np.mean(iou_total)


    return iou_mean

class SEcm_Actor(BaseActor):
    """ Actor for training the our scale estimation module"""

    def delta2xyxy(self, delta):
        '''delta的size是(batch,4)'''
        bbox_cxcywh = delta.clone()
        '''以位于中心的大小为(128,128)的框为基准'''
        bbox_cxcywh[:, :2] = 128.0 + delta[:, :2] * 128.0  # 中心偏移
        bbox_cxcywh[:, 2:] = 128.0 * torch.exp(delta[:, 2:])  # 宽高修正
        '''调整为(x1,y1,x2,y2)格式'''
        bbox_xyxy = bbox_cxcywh.clone()
        bbox_xyxy[:, :2] = bbox_cxcywh[:, :2] - 0.5 * bbox_cxcywh[:, 2:]
        bbox_xyxy[:, 2:] = bbox_cxcywh[:, :2] + 0.5 * bbox_cxcywh[:, 2:]
        return bbox_xyxy

    def __call__(self, data):
        """
        args:
            data - The input data, should contain the fields 'train_images', 'test_images', 'train_anno',
                    'test_proposals' and 'proposal_iou'.

        returns:
            loss    - the training loss
            states  -  dict containing detailed losses
        """
        # Run network to obtain bbox prediction for each test image'
        '''get prediction'''
        pred_dict = self.net(data['train_images'], data['test_images'], data['train_anno'])
        corner_pred = pred_dict['corner']  # (batch,4)
        mask_pred = pred_dict['mask']  # (batch,1,256,256)

        '''get groundtruth'''
        bbox_gt = data['test_anno'].squeeze(0)  # 测试帧的真值框在裁剪出的搜索区域上的坐标(x1,y1,w,h)

        bbox_gt_xyxy = bbox_gt.clone()
        bbox_gt_xyxy[:, 2:] += bbox_gt_xyxy[:, :2]  # (x1,y1,x2,y2)格式

        '''get loss function'''
        mask_loss_F = self.objective['mask']
        corner_loss_F = self.objective['corner']

        '''Compute loss for corner'''
        loss_corner = corner_loss_F(corner_pred, bbox_gt_xyxy)
        if isinstance(loss_corner, tuple):
            loss_corner, iou_corner = loss_corner[0], loss_corner[1].item()
            mask_weight = 1
        else:
            mask_weight = 1000
            iou_corner = iou_cal(corner_pred, bbox_gt_xyxy, wh= False )

        '''Compute total loss'''
        mask_flag = (data['mask'] == 1)  # data[mask]是一个tensor,有batch个元素,其中有的是0有的是1. mask_flag是一个bool型tensor
        num_mask_sample = mask_flag.sum()
        if num_mask_sample > 0:
            mask_gt = data['test_masks'].squeeze(0)  # 测试帧的mask真值框 (batch,1,H,W)
            '''Compute loss for mask'''
            loss_mask = mask_loss_F(mask_pred[mask_flag], mask_gt[mask_flag])  # 只计算那些mask_flag等于1的样本的loss_mask
            mask_weight = mask_weight * loss_mask
            loss = 1 * loss_corner + mask_weight
        else:
            loss_mask = torch.zeros((1,))
            loss = 1 * loss_corner
            mask_weight = loss_corner


        stats = {
            'Loss/total': loss.item(),
            'loss_mask': loss_mask.item(),
            'loss_mask_weight': mask_weight.item(),
            'loss_corner': loss_corner.item(),
            'iou_corner': iou_corner
        }

        return loss, stats
