from . import BaseActor
import torch
import numpy as np
import torch.nn.functional as F

from ltr.train_settings.anchor.anchor import Anchors

def get_cls_loss(pred, label, select):
    if len(select.size()) == 0 or \
            select.size() == torch.Size([0]):
        return 0
    pred = torch.index_select(pred, 0, select)
    label = torch.index_select(label, 0, select)
    return F.nll_loss(pred, label)


def select_cross_entropy_loss(pred, label):
    pred = pred.view(-1, 2)
    label = label.view(-1)
    pos = label.data.eq(1).nonzero().squeeze().cuda()
    neg = label.data.eq(0).nonzero().squeeze().cuda()
    loss_pos = get_cls_loss(pred, label, pos)
    loss_neg = get_cls_loss(pred, label, neg)
    return loss_pos * 0.5 + loss_neg * 0.5


def weight_l1_loss(pred_loc, label_loc, loss_weight):
    b, _, sh, sw = pred_loc.size()
    pred_loc = pred_loc.view(b, 4, -1, sh, sw)  # [b, 4, 5, h, w]
    diff = (pred_loc - label_loc).abs()
    diff = diff.sum(dim=1).view(b, -1, sh, sw)
    loss = diff * loss_weight
    return loss.sum().div(b)


class SiamRPNppActor(BaseActor):

    """ Actor for training the TransT"""
    def __call__(self, data):
        """
        args:
            data - The input data, should contain the fields 'search_images', 'template_images', 'search_anno'.

        returns:
            loss    - the training loss
            states  -  dict containing detailed losses
        """


        # generate labels
        bbox_gt = data['search_anno']
        B, _ = bbox_gt.shape
        bbox_gt_np = bbox_gt.cpu().numpy()

        cls= []
        delta= []
        delta_weight= []
        overlap= []
        for i in range(B):
            cls1, delta1, delta_weight1, overlap1 = self.anchor(target=bbox_gt_np[0,], size=self.anchor.OUTPUT_SIZE, neg=False)
            cls.append(torch.from_numpy(cls1).unsqueeze(0))
            delta.append(torch.from_numpy(delta1).unsqueeze(0))
            delta_weight.append(torch.from_numpy(delta_weight1).unsqueeze(0))
            overlap.append(torch.from_numpy(overlap1).unsqueeze(0))

        cls = torch.cat(cls, dim=0)
        delta = torch.cat(delta, dim=0)
        delta_weight = torch.cat(delta_weight, dim=0)
        overlap = torch.cat(overlap, dim=0)

        weight_dict = self.objective.weight_dict
        input_data = {
            'label_cls': cls.cuda(),
            'label_loc': delta.cuda(),
            'label_loc_weight': delta_weight.cuda(),
            'cls_weight': weight_dict['cls_weight'],
            'loc_weight': weight_dict['loc_weight'],
        }
        cls_pred, loc_pred = self.net(data['search_images'], data['template_images'], input_data)

        cls_loss = select_cross_entropy_loss(cls_pred, input_data['label_cls'])
        loc_loss = weight_l1_loss(loc_pred, input_data['label_loc'], input_data['label_loc_weight'])

        losses = weight_dict['cls_weight']*cls_loss  + weight_dict['loc_weight']*loc_loss

        # Return training stats
        stats = {'Loss/total': losses.item(),
                 'cls_loss/cross_entrophy': cls_loss.item(),
                 'loc_loss/l1': loc_loss.item(),
                 }

        return losses, stats


class BlockActor_cls(BaseActor):
    """ Actor for training the TransT"""
    def __call__(self, data):
        """
        args:
            data - The input data, should contain the fields 'search_images', 'template_images', 'search_anno'.

        returns:
            loss    - the training loss
            states  -  dict containing detailed losses
        """
        import time
        outputs = self.net(data['search_images'], data['template_images'])

        # generate labels
        targets =[]
        targets_origin = data['search_anno']
        for i in range(len(targets_origin)):
            h, w =data['search_images'][i][0].shape
            target_origin = targets_origin[i]
            target = {}
            target_origin = target_origin.reshape([1,-1])
            target_origin[0][0] += target_origin[0][2] / 2
            target_origin[0][0] /= w
            target_origin[0][1] += target_origin[0][3] / 2
            target_origin[0][1] /= h
            target_origin[0][2] /= w
            target_origin[0][3] /= h
            target['boxes'] = target_origin
            label = np.array([0])
            label = torch.tensor(label, device=data['search_anno'].device)
            target['labels'] = label
            targets.append(target)

        # Compute loss
        # outputs:(center_x, center_y, width, height)
        if isinstance(self.objective, dict):
            loss_dict = self.objective['tracking'](outputs, targets)
            weight_dict = self.objective['tracking'].weight_dict

        else:
            loss_dict = self.objective(outputs, targets)
            weight_dict = self.objective.weight_dict

        losses = sum(loss_dict[k] * weight_dict[k] for k in loss_dict.keys() if k in weight_dict)

        # Return training stats
        stats = {'cls_only_Loss/total': losses.item(),
                 'Loss/ce': loss_dict['loss_ce'].item(),
                 }

        return losses, stats
