from . import BaseActor
import torch
import numpy as np


def mask_iou(pred, target):

    """
    param: pred of size [N x H x W]
    param: target of size [N x H x W]
    """

    assert len(pred.shape) == 3 and pred.shape == target.shape

    N = pred.size(0)

    inter = torch.min(pred, target).sum(2).sum(1)
    union = torch.max(pred, target).sum(2).sum(1)

    iou = torch.sum(inter / union) / N

    return iou


def mask_iou_loss(pred, mask, num_object):

    N, K, H, W = mask.shape
    loss = torch.zeros(1).to(pred.device)
    start = 0 if K == num_object else 1

    iou_total = []
    for i in range(N):
        iou_i = mask_iou(pred[i, start:num_object+start], mask[i, start:num_object+start])
        loss += (1.0 - iou_i)
        iou_total = iou_i.clone().detach().cpu().numpy()


    loss = loss / N
    mean_iou = np.mean(iou_total)
    return loss, mean_iou


class TrackSegActor(BaseActor):
    """ Actor for training the TransT"""
    def __call__(self, data):
        """
        args:
            data - The input data, should contain the fields 'search_images', 'template_images', 'search_anno'.

        returns:
            loss    - the training loss
            states  -  dict containing detailed losses
            (data['search_images'], data['template_images'])
            search -> test
            template -> train
        """
        outputs = self.net(data['search_images'], data['template_images'])

        # generate labels
        targets = []
        targets_origin = data['search_anno']
        for i in range(len(targets_origin)):
            h, w = data['search_images'][i][0].shape
            target_origin = targets_origin[i]
            target = {}
            target_origin = target_origin.reshape([1, -1])
            target_origin[0][0] += target_origin[0][2] / 2
            target_origin[0][0] /= w
            target_origin[0][1] += target_origin[0][3] / 2
            target_origin[0][1] /= h
            target_origin[0][2] /= w
            target_origin[0][3] /= h
            target['boxes'] = target_origin
            label = np.array([0])
            label = torch.tensor(label, device=data['search_anno'].device)
            target['labels'] = label
            targets.append(target)

        # Compute loss
        # outputs:(center_x, center_y, width, height)
        loss_dict = self.objective['tracking'](outputs, targets)
        weight_dict = self.objective['tracking'].weight_dict
        losses_tracking = sum(loss_dict[k] * weight_dict[k] for k in loss_dict.keys() if k in weight_dict)

        mask_loss_F = self.objective['segment']

        """extra mask iou loss"""
        '''Compute total loss'''
        mask_pred = outputs['mask_predict']  # (batch,1,256,256)
        mask_flag = (data['mask'] == 1)  # data[mask]是一个tensor,有batch个元素,其中有的是0有的是1. mask_flag是一个bool型tensor
        num_mask_sample = mask_flag.sum()
        mask_weight_b = 4#6
        mask_weight_iou = 6 #8

        if num_mask_sample > 0:
            mask_gt = data['search_masks'].unsqueeze(1) # 测试帧的mask真值框 (batch,1,H,W)
            '''Compute loss for mask'''
            loss_mask = mask_loss_F(mask_pred[mask_flag], mask_gt[mask_flag])  # 只计算那些mask_flag等于1的样本的loss_mask
            loss_mask_iou, mean_iou = mask_iou_loss(mask_pred[mask_flag], mask_gt[mask_flag], 1)
            losses_segment = mask_weight_b * loss_mask + mask_weight_iou * loss_mask_iou
            losses = 1 * losses_tracking + losses_segment
            stats = {'Loss/total': losses.item(),

                     'Loss/tracking': losses_tracking.item(),
                     'Loss/segment': losses_segment.item(),

                     'Loss/ce': loss_dict['loss_ce'].item(),
                     'Loss/bbox': loss_dict['loss_bbox'].item(),
                     'Loss/giou': loss_dict['loss_giou'].item(),

                     'Loss/mask_binary': loss_mask.item(),
                     'Loss/mask_iou': loss_mask_iou.item(),

                     'maskiou': mean_iou.item(),
                     'iou': loss_dict['iou'].item(),
                     }
        else:

            losses = losses_tracking
            loss_mask = 0
            mean_iou =0
            losses_segment=0
            loss_mask_iou=0

            stats = {'Loss/total': losses.item(),

                     'Loss/tracking': losses_tracking.item(),
                     'Loss/segment': losses_segment,

                     'Loss/ce': loss_dict['loss_ce'].item(),
                     'Loss/bbox': loss_dict['loss_bbox'].item(),
                     'Loss/giou': loss_dict['loss_giou'].item(),

                     'Loss/mask_binary': loss_mask,
                     'Loss/mask_iou': loss_mask_iou,

                     'maskiou': mean_iou,
                     'iou': loss_dict['iou'].item(),
                     }

        return losses, stats
