
from __future__ import absolute_import

import numpy as np
import math
import torchvision.transforms.functional as tvisf
import cv2
import torch
import torch.nn.functional as F
import time
from pysot_toolkit.trackers.vot20_utils import *
import pysot_toolkit.vot2020 as vot

from pytracking.refine_modules.refine_module_vot20 import RefineModule

import os
from pysot_toolkit.trackers.net_wrappers import NetWithBackbone
import random

class TrackingNC_ALPHA(object):

    def __init__(self, name, net, window_penalty=0.49, exemplar_size=128, instance_size=256,
                 refine_path='/tmp/code/pysot_toolkit/ar.pth.tar',search_z=3.0, search_x=3.0):
        self.name = name
        self.net = net
        self.window_penalty = window_penalty
        self.exemplar_size = exemplar_size
        self.instance_size = instance_size
        self.net = self.net

        self.alpha = RefineModule(refine_path, 2, 256)
        self.THRES = 0.65

        self.search_z = search_z
        self.search_x = search_x

    def _convert_score(self, score):

        score = score.permute(2, 1, 0).contiguous().view(2, -1).permute(1, 0)
        score = F.softmax(score, dim=1).data[:, 0].cpu().numpy()
        return score

    def _convert_bbox(self, delta):

        delta = delta.permute(2, 1, 0).contiguous().view(4, -1)
        delta = delta.data.cpu().numpy()

        return delta

    def _bbox_clip(self, cx, cy, width, height, boundary):
        cx = max(0, min(cx, boundary[1]))
        cy = max(0, min(cy, boundary[0]))
        width = max(10, min(width, boundary[1]))
        height = max(10, min(height, boundary[0]))
        return cx, cy, width, height

    def _bbox_clip_xyxy(self, x1, y1, x2, y2, boundary, min_sz=10):
        '''boundary (H,W)'''
        x1_new = max(0, min(x1, boundary[1] - min_sz))
        y1_new = max(0, min(y1, boundary[0] - min_sz))
        x2_new = max(min_sz, min(x2, boundary[1]))
        y2_new = max(min_sz, min(y2, boundary[0]))
        return x1_new, y1_new, x2_new, y2_new

    def get_subwindow(self, im, pos, model_sz, original_sz, avg_chans):
        """
        args:
            im: rgb based image
            pos: center position
            model_sz: exemplar size
            original_sz: original size
            avg_chans: channel average
        """
        if isinstance(pos, float):
            pos = [pos, pos]
        sz = original_sz
        im_sz = im.shape
        c = (original_sz + 1) / 2
        # context_xmin = round(pos[0] - c) # py2 and py3 round
        context_xmin = np.floor(pos[0] - c + 0.5)
        context_xmax = context_xmin + sz - 1
        # context_ymin = round(pos[1] - c)
        context_ymin = np.floor(pos[1] - c + 0.5)
        context_ymax = context_ymin + sz - 1
        left_pad = int(max(0., -context_xmin))
        top_pad = int(max(0., -context_ymin))
        right_pad = int(max(0., context_xmax - im_sz[1] + 1))
        bottom_pad = int(max(0., context_ymax - im_sz[0] + 1))

        context_xmin = context_xmin + left_pad
        context_xmax = context_xmax + left_pad
        context_ymin = context_ymin + top_pad
        context_ymax = context_ymax + top_pad

        r, c, k = im.shape
        if any([top_pad, bottom_pad, left_pad, right_pad]):
            size = (r + top_pad + bottom_pad, c + left_pad + right_pad, k)
            te_im = np.zeros(size, np.uint8)
            te_im[top_pad:top_pad + r, left_pad:left_pad + c, :] = im
            if top_pad:
                te_im[0:top_pad, left_pad:left_pad + c, :] = avg_chans
            if bottom_pad:
                te_im[r + top_pad:, left_pad:left_pad + c, :] = avg_chans
            if left_pad:
                te_im[:, 0:left_pad, :] = avg_chans
            if right_pad:
                te_im[:, c + left_pad:, :] = avg_chans
            im_patch = te_im[int(context_ymin):int(context_ymax + 1),
                       int(context_xmin):int(context_xmax + 1), :]
        else:
            im_patch = im[int(context_ymin):int(context_ymax + 1),
                       int(context_xmin):int(context_xmax + 1), :]

        if not np.array_equal(model_sz, original_sz):
            im_patch = cv2.resize(im_patch, (model_sz, model_sz))
        im_patch = im_patch.transpose(2, 0, 1)
        im_patch = im_patch[np.newaxis, :, :, :]
        im_patch = im_patch.astype(np.float32)
        im_patch = torch.from_numpy(im_patch)
        im_patch = im_patch.cuda()
        return im_patch

    def initialize_features(self):
        if not getattr(self, 'features_initialized', False):
            self.net.initialize()
        self.features_initialized = True

    def initialize(self, image, mask):
        tic = time.time()
        hanning = np.hanning(14)  # 14
        window = np.outer(hanning, hanning)
        self.window = window.flatten()
        # Initialize
        self.initialize_features()
        """"""""""""
        if isinstance(mask, list):
            region = mask
        else:
            region = rect_from_mask(mask)

            self.first_mask = mask[region[1]: region[1] + region[3], region[0]:region[0]+region[2]]


        self.H, self.W, _ = image.shape
        gt_bbox_np = np.array(region).astype(np.float32)


        '''initilize refinement module for specific video'''
        self.alpha.initialize(image, gt_bbox_np)

        bbox = gt_bbox_np
        self.center_pos = np.array([bbox[0] + bbox[2] / 2,
                                    bbox[1] + bbox[3] / 2])
        self.size = np.array([bbox[2], bbox[3]])

        # calculate z crop size
        w_z = self.size[0] + (self.search_z - 1) * ((self.size[0] + self.size[1]) * 0.5)  #
        h_z = self.size[1] + (self.search_z - 1) * ((self.size[0] + self.size[1]) * 0.5)
        s_z = math.ceil(math.sqrt(w_z * h_z))

        # calculate channle average
        self.channel_average = np.mean(image, axis=(0, 1))

        # get crop
        z_crop = self.get_subwindow(image, self.center_pos,
                                    self.exemplar_size,
                                    s_z, self.channel_average)

        # normalize
        z_crop = z_crop.float().mul(1.0 / 255.0).clamp(0.0, 1.0)
        self.mean = [0.485, 0.456, 0.406]
        self.std = [0.229, 0.224, 0.225]
        self.inplace = False
        z_crop[0] = tvisf.normalize(z_crop[0], self.mean, self.std, self.inplace)

        # initialize template feature
        self.net.template(z_crop)
        out = {'time': time.time() - tic}
        return out

    def track(self, image, info=None):
        # calculate x crop size
        w_x = self.size[0] + (self.search_x - 1) * ((self.size[0] + self.size[1]) * 0.5)
        h_x = self.size[1] + (self.search_x - 1) * ((self.size[0] + self.size[1]) * 0.5)
        s_x = math.ceil(math.sqrt(w_x * h_x))

        # get crop
        x_crop = self.get_subwindow(image, self.center_pos,
                                    self.instance_size,
                                    round(s_x), self.channel_average)

        # normalize
        x_crop = x_crop.float().mul(1.0 / 255.0).clamp(0.0, 1.0)
        x_crop[0] = tvisf.normalize(x_crop[0], self.mean, self.std, self.inplace)

        # track
        outputs = self.net.track(x_crop)
        score = self._convert_score(outputs['pred_logits'])
        pred_bbox = self._convert_bbox(outputs['pred_boxes'])

        # window penalty
        pscore = score * (1 - self.window_penalty) + \
                 self.window * self.window_penalty

        best_idx = np.argmax(pscore)
        bbox = pred_bbox[:, best_idx]
        bbox = bbox * s_x
        cx = bbox[0] + self.center_pos[0] - s_x / 2
        cy = bbox[1] + self.center_pos[1] - s_x / 2
        width = bbox[2]
        height = bbox[3]

        # clip boundary
        cx, cy, width, height = self._bbox_clip(cx, cy, width,
                                                height, image.shape[:2])

        # update state
        self.center_pos = np.array([cx, cy])
        self.size = np.array([width, height])

        bbox = [cx - width / 2,
                cy - height / 2,
                width,
                height]

        out = {'target_bbox': bbox,
               'best_score': pscore}

        '''Mask'''
        x1, y1, w, h = bbox
        x1_n, y1_n, x2_n, y2_n = self._bbox_clip_xyxy(x1, y1, x1 + w, y1 + h, (self.H, self.W))
        w_n = x2_n - x1_n
        h_n = y2_n - y1_n
        bbox_new = [x1_n, y1_n, w_n, h_n]
        pred_mask = self.alpha.refine(image, np.array(bbox_new))
        final_mask = (pred_mask > self.THRES).astype(np.uint8)


        return bbox_new, final_mask

"""
pixs = image.shape[0] * image.shape[1]
        r = (w * h) / pixs
        count = (final_mask == 1).sum()
        if count < (r * 0.08 * pixs):  # 0.05
            a, b, c, d = int(x1_n), int(y1_n), int(w_n), int(h_n)
            e, f = int(w), int(h),
            first_rz = cv2.resize(self.first_mask.copy(), (e, f))
            final_mask[b:b+d, a:a+c] = first_rz[0:d, 0:c]
        else:
            pass
"""

def set_seed(seed):
    random.seed(seed)
    os.environ['PYTHONHASHSEED'] = str(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.backends.cudnn.benchmark = True
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.enabled = True


def run_vot_exp(pth = None):
    VIS = False
    pth = 18
    print(str(pth))

    pth_dict={
        1: '/tmp/code/pytracking/networks/CrossViT_ep0080.pth.tar',
        2: '/tmp/code/pytracking/networks/CrossViT_ep0081.pth.tar',
        3: '/tmp/code/pytracking/networks/CrossViT_ep0082.pth.tar',
        4: '/tmp/code/pytracking/networks/CrossViT_ep0083.pth.tar',
        5: '/tmp/code/pytracking/networks/CrossViT_ep0084.pth.tar',
        6: '/tmp/code/pytracking/networks/CrossViT_ep0085.pth.tar',
        7: '/tmp/code/pytracking/networks/CrossViT_ep0086.pth.tar',
        8: '/tmp/code/pytracking/networks/CrossViT_ep0087.pth.tar',
        9: '/tmp/code/pytracking/networks/CrossViT_ep0088.pth.tar',
        10: '/tmp/code/pytracking/networks/CrossViT_ep0089.pth.tar',
        11: '/tmp/code/pytracking/networks/CrossViT_ep0090.pth.tar',
        12: '/tmp/code/pytracking/networks/CrossViT_ep0091.pth.tar',
        13: '/tmp/code/pytracking/networks/CrossViT_ep0092.pth.tar',
        14: '/tmp/code/pytracking/networks/CrossViT_ep0093.pth.tar',
        15: '/tmp/code/pytracking/networks/CrossViT_ep0094.pth.tar',
        16: '/tmp/code/pytracking/networks/CrossViT_ep0095.pth.tar',
        17: '/tmp/code/pytracking/networks/CrossViT_ep0096.pth.tar',
        18: '/home/v-xfei/Public/project/TransT/pytracking/networks/CrossViT_ep0100.pth.tar',
    }
    #########################################
    """
    import json
    save_path = '/tmp/code/pt_dirs1/hp_rand.json'
    with open(save_path, encoding='utf8') as fp:
        hp = json.load(fp)
    """

    #net_path = ' '#pth_dict[pth]
    net_path = '/home/v-xfei/Public/project/TransT/pytracking/networks/CrossViT_ep0051.pth.tar'

    print(net_path)

    seed_test = 1000
    set_seed(seed_test)

    torch.set_num_threads(1)
    # torch.cuda.set_device(CUDA_ID)  # set GPU id
    save_root = '/home/v-xfei/Public/dataset/vot_tools/vot21/vis/'
    if VIS and (not os.path.exists(save_root)):
        os.mkdir(save_root)
    # Absolute path of the model
    # create model
    net = NetWithBackbone(net_path=net_path, use_gpu=True)
    tracker = TrackingNC_ALPHA(name='dualtfr160', net=net, window_penalty=0.49, exemplar_size=224, instance_size=224, search_z=3, search_x=3.0,
                               refine_path='/home/v-xfei/Public/project/TransT/pysot_toolkit/ar.pth.tar')

    handle = vot.VOT("mask")
    selection = handle.region()
    imagefile = handle.frame()

    if VIS:
        '''for vis'''
        seq_name = imagefile.split('/')[-3]
        save_v_dir = os.path.join(save_root,seq_name)
        if not os.path.exists(save_v_dir):
            os.mkdir(save_v_dir)
        cur_time = int(time.time() % 10000)
        save_dir = os.path.join(save_v_dir, str(cur_time))
        if not os.path.exists(save_dir):
            os.mkdir(save_dir)

    image = cv2.cvtColor(cv2.imread(imagefile), cv2.COLOR_BGR2RGB) # Right
    # mask given by the toolkit ends with the target (zero-padding to the right and down is needed)
    mask = make_full_size(selection, (image.shape[1], image.shape[0]))
    tracker.initialize(image, mask)

    while True:
        imagefile = handle.frame()
        if not imagefile:
            break
        image = cv2.cvtColor(cv2.imread(imagefile), cv2.COLOR_BGR2RGB)  # Right
        b1, m = tracker.track(image)
        handle.report(m)
        if VIS:
            '''Visualization'''
            image_show = image[:,:,::-1].copy() # RGB --> BGR
            image_show = image_show.astype(np.float32)

            image_show[:,:,0] += 127.0 * m
            image_show = image_show.clip(0,255).astype(np.uint8)
            cv2.rectangle(image_show,(int(b1[0]),int(b1[1])),
                          (int(b1[0]+b1[2]),int(b1[1]+b1[3])),(0,0,255),2)
            save_path = os.path.join(save_dir, imagefile.split('/')[-1])
            cv2.imwrite(save_path,image_show)
