
from __future__ import absolute_import

import numpy as np
import math
import torchvision.transforms.functional as tvisf
import cv2
import torch
import torch.nn.functional as F
import time
from pysot_toolkit.trackers.vot20_utils import *
import pysot_toolkit.vot2020 as vot

from pytracking.refine_modules.refine_module_vot20 import RefineModule

import os
from pysot_toolkit.trackers.net_wrappers import NetWithBackbone

def mask_torch2numpy(Pmask):
    Pmask_arr = np.array(Pmask.squeeze().cpu())  # (H,W) (0,1)
    return Pmask_arr

def map_mask_back(im, target_bb, search_area_factor, mask, mode=cv2.BORDER_REPLICATE):
    """ Extracts a crop centered at target_bb box, of size search_area_factor times target_bb(Both height and width)

    args:
        im - cv image
        target_bb - target box [x, y, w, h]
        search_area_factor - Ratio of crop size to target size
        output_sz - (float) Size to which the extracted crop is resized (always square). If None, no resizing is done.

    returns:
        cv image - extracted crop
        float - the factor by which the crop has been resized to make the crop size equal output_size
    """
    H,W = (im.shape[0],im.shape[1])
    base = np.zeros((H,W))
    x, y, w, h = target_bb.tolist()

    # Crop image
    #ws = math.ceil(search_area_factor * w)
    #hs = math.ceil(search_area_factor * h)
    ws = w + (search_area_factor - 1) * ((w + h) * 0.5)
    hs = h + (search_area_factor - 1) * ((w + h) * 0.5)

    s_x = math.ceil(math.sqrt(ws * hs))

    ws = s_x
    hs = s_x

    if ws < 1 or hs < 1:
        raise Exception('Too small bounding box.')

    x1 = round(x + 0.5*w - ws*0.5)
    x2 = x1 + ws

    y1 = round(y + 0.5 * h - hs * 0.5)
    y2 = y1 + hs

    x1_pad = max(0, -x1)
    x2_pad = max(x2-im.shape[1]+1, 0)

    y1_pad = max(0, -y1)
    y2_pad = max(y2-im.shape[0]+1, 0)

    '''pad base'''
    base_padded = cv2.copyMakeBorder(base, y1_pad, y2_pad, x1_pad, x2_pad, mode)
    '''Resize mask'''
    mask_rsz = cv2.resize(mask,(ws,hs))
    '''fill region with mask'''
    base_padded[y1+y1_pad:y2+y1_pad, x1+x1_pad:x2+x1_pad] = mask_rsz.copy()
    '''crop base_padded to get final mask'''
    final_mask = base_padded[y1_pad:y1_pad+H,x1_pad:x1_pad+W]
    assert (final_mask.shape == (H,W))
    return final_mask

def map_mask_reset(im, target_bb, mask, mode=cv2.BORDER_REPLICATE):

    H,W = (im.shape[0],im.shape[1])
    base = np.zeros((H,W))
    x, y, w, h = target_bb.tolist()

    ws = int(w)
    hs = int(h)

    x1 = round(x + 0.5*w - ws*0.5)
    x2 = x1 + ws

    y1 = round(y + 0.5 * h - hs * 0.5)
    y2 = y1 + hs

    x1_pad = max(0, -x1)
    x2_pad = max(x2-im.shape[1]+1, 0)

    y1_pad = max(0, -y1)
    y2_pad = max(y2-im.shape[0]+1, 0)

    '''pad base'''
    base_padded = cv2.copyMakeBorder(base, y1_pad, y2_pad, x1_pad, x2_pad, mode)
    '''Resize mask'''
    mask_rsz = cv2.resize(mask,(ws,hs))
    '''fill region with mask'''
    base_padded[y1+y1_pad:y2+y1_pad, x1+x1_pad:x2+x1_pad] = mask_rsz.copy()
    '''crop base_padded to get final mask'''
    final_mask = base_padded[y1_pad:y1_pad+H,x1_pad:x1_pad+W]
    assert (final_mask.shape == (H,W))
    return final_mask

def iou(box1, box2, wh=True):
    """
    compute the iou of two boxes.
    Args:
        box1, box2: [xmin, ymin, xmax, ymax] (wh=False) or [xcenter, ycenter, w, h] (wh=True)
        wh: the format of coordinate.
    Return:
        iou: iou of box1 and box2.
    """
    if wh == False:
        xmin1, ymin1, xmax1, ymax1 = box1
        xmin2, ymin2, xmax2, ymax2 = box2
    else:
        xmin1, ymin1 = int(box1[0] - box1[2] / 2.0), int(box1[1] - box1[3] / 2.0)
        xmax1, ymax1 = int(box1[0] + box1[2] / 2.0), int(box1[1] + box1[3] / 2.0)
        xmin2, ymin2 = int(box2[0] - box2[2] / 2.0), int(box2[1] - box2[3] / 2.0)
        xmax2, ymax2 = int(box2[0] + box2[2] / 2.0), int(box2[1] + box2[3] / 2.0)

    xx1 = np.max([xmin1, xmin2])
    yy1 = np.max([ymin1, ymin2])
    xx2 = np.min([xmax1, xmax2])
    yy2 = np.min([ymax1, ymax2])

    area1 = (xmax1 - xmin1) * (ymax1 - ymin1)
    area2 = (xmax2 - xmin2) * (ymax2 - ymin2)

    inter_area = (np.max([0, xx2 - xx1])) * (np.max([0, yy2 - yy1]))
    iou = inter_area / (area1 + area2 - inter_area + 1e-6)

    return iou

class TrackingNC_ALPHA(object):

    def __init__(self, name, net, window_penalty=0.49, exemplar_size=224, instance_size=224,
                 refine_path='/home/v-xfei/Public/project/TransT/pysot_toolkit/ar.pth.tar',search_z=3.0, search_x=3.0):
        self.name = name
        self.net = net
        self.window_penalty = window_penalty
        self.exemplar_size = exemplar_size
        self.instance_size = instance_size
        self.net = self.net

        self.alpha = RefineModule(refine_path, 2, 256)

        self.search_z = search_z
        self.search_x = search_x
        self.search_mask = 2

        self.THRES = 0.50 # 0.65


    def _convert_score(self, score):

        score = score.permute(2, 1, 0).contiguous().view(2, -1).permute(1, 0)
        score = F.softmax(score, dim=1).data[:, 0].cpu().numpy()
        return score

    def _convert_bbox(self, delta):

        delta = delta.permute(2, 1, 0).contiguous().view(4, -1)
        delta = delta.data.cpu().numpy()

        return delta

    def _bbox_clip(self, cx, cy, width, height, boundary):
        cx = max(0, min(cx, boundary[1]))
        cy = max(0, min(cy, boundary[0]))
        width = max(10, min(width, boundary[1]))
        height = max(10, min(height, boundary[0]))
        return cx, cy, width, height

    def get_subwindow(self, im, pos, model_sz, original_sz, avg_chans):
        """
        args:
            im: rgb based image
            pos: center position
            model_sz: exemplar size
            original_sz: original size
            avg_chans: channel average
        """
        if isinstance(pos, float):
            pos = [pos, pos]
        sz = original_sz
        im_sz = im.shape
        c = (original_sz + 1) / 2
        # context_xmin = round(pos[0] - c) # py2 and py3 round
        context_xmin = np.floor(pos[0] - c + 0.5)
        context_xmax = context_xmin + sz - 1
        # context_ymin = round(pos[1] - c)
        context_ymin = np.floor(pos[1] - c + 0.5)
        context_ymax = context_ymin + sz - 1
        left_pad = int(max(0., -context_xmin))
        top_pad = int(max(0., -context_ymin))
        right_pad = int(max(0., context_xmax - im_sz[1] + 1))
        bottom_pad = int(max(0., context_ymax - im_sz[0] + 1))

        context_xmin = context_xmin + left_pad
        context_xmax = context_xmax + left_pad
        context_ymin = context_ymin + top_pad
        context_ymax = context_ymax + top_pad

        r, c, k = im.shape
        if any([top_pad, bottom_pad, left_pad, right_pad]):
            size = (r + top_pad + bottom_pad, c + left_pad + right_pad, k)
            te_im = np.zeros(size, np.uint8)
            te_im[top_pad:top_pad + r, left_pad:left_pad + c, :] = im
            if top_pad:
                te_im[0:top_pad, left_pad:left_pad + c, :] = avg_chans
            if bottom_pad:
                te_im[r + top_pad:, left_pad:left_pad + c, :] = avg_chans
            if left_pad:
                te_im[:, 0:left_pad, :] = avg_chans
            if right_pad:
                te_im[:, c + left_pad:, :] = avg_chans
            im_patch = te_im[int(context_ymin):int(context_ymax + 1),
                       int(context_xmin):int(context_xmax + 1), :]
        else:
            im_patch = im[int(context_ymin):int(context_ymax + 1),
                       int(context_xmin):int(context_xmax + 1), :]

        if not np.array_equal(model_sz, original_sz):
            im_patch = cv2.resize(im_patch, (model_sz, model_sz))
        im_patch = im_patch.transpose(2, 0, 1)
        im_patch = im_patch[np.newaxis, :, :, :]
        im_patch = im_patch.astype(np.float32)
        im_patch = torch.from_numpy(im_patch)
        im_patch = im_patch.cuda()
        return im_patch

    def initialize_features(self):
        if not getattr(self, 'features_initialized', False):
            self.net.initialize()
        self.features_initialized = True

    def initialize(self, image, mask):
        tic = time.time()
        hanning = np.hanning(14)  # 14
        window = np.outer(hanning, hanning)
        self.window = window.flatten()
        # Initialize
        self.initialize_features()
        """"""""""""
        if isinstance(mask, list):
            region = mask
        elif isinstance(mask, dict):
            region = mask['init_bbox']
           # self.prev_mask_region = np.ones([int(region[3]), int(region[2])])
        else:
            region = rect_from_mask(mask)
            #self.prev_mask_region = mask[int(region[1]):int(region[1] + region[3]), int(region[0]):int(region[0] + region[2])]

        self.H, self.W, _ = image.shape
        gt_bbox_np = np.array(region).astype(np.float32)


        '''initilize refinement module for specific video'''
        self.alpha.initialize(image, gt_bbox_np)

        bbox = gt_bbox_np
        self.center_pos = np.array([bbox[0] + bbox[2] / 2,
                                    bbox[1] + bbox[3] / 2])
        self.size = np.array([bbox[2], bbox[3]])

        # calculate z crop size
        w_z = self.size[0] + (self.search_z - 1) * ((self.size[0] + self.size[1]) * 0.5)  #
        h_z = self.size[1] + (self.search_z - 1) * ((self.size[0] + self.size[1]) * 0.5)
        s_z = math.ceil(math.sqrt(w_z * h_z))

        # calculate channle average
        self.channel_average = np.mean(image, axis=(0, 1))

        # get crop
        z_crop = self.get_subwindow(image, self.center_pos,
                                    self.exemplar_size,
                                    s_z, self.channel_average)

        # normalize
        z_crop = z_crop.float().mul(1.0 / 255.0).clamp(0.0, 1.0)
        self.mean = [0.485, 0.456, 0.406]
        self.std = [0.229, 0.224, 0.225]
        self.inplace = False
        z_crop[0] = tvisf.normalize(z_crop[0], self.mean, self.std, self.inplace)

        # initialize template feature
        self.net.template(z_crop)
        out = {'time': time.time() - tic}
        self.frame_num=1
        return out

    def track(self, image, info=None):
        self.frame_num+=1
        self.bbox_prev = [self.center_pos[0] - self.size[0]*0.5,
                          self.center_pos[1] - self.size[1]*0.5, self.size[0], self.size[1]]


        # calculate x crop size
        w_x = self.size[0] + (self.search_x - 1) * ((self.size[0] + self.size[1]) * 0.5)
        h_x = self.size[1] + (self.search_x - 1) * ((self.size[0] + self.size[1]) * 0.5)
        s_x = math.ceil(math.sqrt(w_x * h_x))

        # get crop
        x_crop = self.get_subwindow(image, self.center_pos,
                                    self.instance_size,
                                    round(s_x), self.channel_average)

        # normalize
        x_crop = x_crop.float().mul(1.0 / 255.0).clamp(0.0, 1.0)
        x_crop[0] = tvisf.normalize(x_crop[0], self.mean, self.std, self.inplace)

        # track
        outputs = self.net.track(x_crop)
        score = self._convert_score(outputs['pred_logits'])
        pred_bbox = self._convert_bbox(outputs['pred_boxes'])

        # window penalty
        pscore = score * (1 - self.window_penalty) + \
                 self.window * self.window_penalty

        best_idx = np.argmax(pscore)
        bbox = pred_bbox[:, best_idx]
        bbox = bbox * s_x
        cx = bbox[0] + self.center_pos[0] - s_x / 2
        cy = bbox[1] + self.center_pos[1] - s_x / 2
        width = bbox[2]
        height = bbox[3]

        # clip boundary
        cx, cy, width, height = self._bbox_clip(cx, cy, width,
                                                height, image.shape[:2])

        # update state
        self.center_pos = np.array([cx, cy])
        self.size = np.array([width, height])

        bbox = [cx - width / 2,
                cy - height / 2,
                width,
                height]

        out = {'target_bbox': bbox,
               'best_score': pscore}

        pred_bbox = out['target_bbox']
        """
        
        """
        ''' Mask report'''
        pred_mask = outputs['mask_predict'][0, 0,].detach().cpu().numpy()
        pred_mask = (pred_mask > self.THRES).astype(np.uint8)
        ''' Mask Post-processing'''
        pbbox = np.array(self.bbox_prev)
        self.cbbox = np.array(pred_bbox)
        ori_mask = map_mask_back(image, pbbox, 3, pred_mask, mode=cv2.BORDER_CONSTANT)
        ori_mask = ori_mask.astype(np.uint8)
        contours, hierarchy = cv2.findContours(ori_mask.copy(), cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
        mask_candi = []
        re_mask_flag = False
        for cont in contours:
            x1 = np.min(cont[:, 0, 0])
            y1 = np.min(cont[:, 0, 1])
            x2 = np.max(cont[:, 0, 0])
            y2 = np.max(cont[:, 0, 1])
            cx_conts = (x1 + x2) * 0.5
            cy_conts = (y1 + y2) * 0.5
            w_conts = x2 - x1
            h_conts = y2 - y1
            contour_candi = [int(x1), int(y1), int(x2), int(y2)]
            if abs(cx_conts - self.center_pos[0]) < self.size[0] and abs(cy_conts - self.center_pos[1]) < self.size[1]:
                if iou(pred_bbox, [x1, y1, w_conts, h_conts]) > 0.55:
                    mask_candi.append(contour_candi)
                if len(mask_candi) > 1:
                    re_mask_flag = True
        # if self.frame_num == 40:
        # a=1
        if len(mask_candi) == 0: re_mask_flag = True
        if re_mask_flag:
            w_x = self.size[0] + (self.search_mask - 1) * ((self.size[0] + self.size[1]) * 0.5)
            h_x = self.size[1] + (self.search_mask - 1) * ((self.size[0] + self.size[1]) * 0.5)
            s_x = math.ceil(math.sqrt(w_x * h_x))
            # get crop
            x_mask_crop = self.get_subwindow(image, self.center_pos,
                                             self.instance_size,
                                             round(s_x), self.channel_average)

            # normalize
            x_crop = x_crop.float().mul(1.0 / 255.0).clamp(0.0, 1.0)
            x_crop[0] = tvisf.normalize(x_crop[0], self.mean, self.std, self.inplace)
            # track
            mask_re_predict = self.net.segm(x_mask_crop)
            mask_re_predict = mask_re_predict[0, 0,].detach().cpu().numpy()
            mask_re_predict = (mask_re_predict > self.THRES).astype(np.uint8)
            final_mask = map_mask_back(image, self.cbbox, self.search_mask, mask_re_predict, mode=cv2.BORDER_CONSTANT)
        #  print('reset: ' + str(self.frame_num))
        else:
            H, W = (image.shape[0], image.shape[1])
            final_mask = np.zeros((H, W))
            b = mask_candi[0]
            final_mask[b[1]:b[3], b[0]:b[2]] = ori_mask[b[1]:b[3], b[0]:b[2]]

        #pred_mask_ar = self.alpha.refine_mask(image, np.array(bbox))
       # final_mask_ar = (pred_mask_ar > self.THRES).astype(np.uint8)
       # final_mask_ar = None
        final_mask = final_mask.astype(np.uint8)
        return pred_bbox, final_mask

def run_vot_exp(pth=None):
    print(str(pth))
    VIS = False
    pth_dict = {
        0: '/home/v-xfei/Public/project/TransT/pytracking/networks/CrossViT_ep0052.pth.tar',
        1: '/home/v-xfei/Public/project/TransT/pytracking/networks/CrossViT_ep0080.pth.tar',
        2: '/home/v-xfei/Public/project/TransT/pytracking/networks/CrossViT_ep0081.pth.tar',
        3: '/home/v-xfei/Public/project/TransT/pytracking/networks/CrossViT_ep0082.pth.tar',
        4: '/home/v-xfei/Public/project/TransT/pytracking/networks/CrossViT_ep0083.pth.tar',
        5: '/home/v-xfei/Public/project/TransT/pytracking/networks/CrossViT_ep0084.pth.tar',
        6: '/home/v-xfei/Public/project/TransT/pytracking/networks/CrossViT_ep0085.pth.tar',
        7: '/tmp/code/pytracking/networks/CrossViT_ep0086.pth.tar',
        8: '/tmp/code/pytracking/networks/CrossViT_ep0087.pth.tar',
        9: '/tmp/code/pytracking/networks/CrossViT_ep0088.pth.tar',
        10: '/tmp/code/pytracking/networks/CrossViT_ep0089.pth.tar',
        11: '/tmp/code/pytracking/networks/CrossViT_ep0090.pth.tar',
        12: '/tmp/code/pytracking/networks/CrossViT_ep0091.pth.tar',
        13: '/tmp/code/pytracking/networks/CrossViT_ep0092.pth.tar',
        14: '/tmp/code/pytracking/networks/CrossViT_ep0093.pth.tar',
        15: '/tmp/code/pytracking/networks/CrossViT_ep0094.pth.tar',
        16: '/tmp/code/pytracking/networks/CrossViT_ep0095.pth.tar',
        17: '/tmp/code/pytracking/networks/CrossViT_ep0096.pth.tar',
        18: '/tmp/code/pytracking/networks/CrossViT_ep0097.pth.tar',
    }
    #########################################
    #if isinstance(pth, int):
    #    net_path = pth_dict[pth]
    net_path = '/home/v-xfei/Public/project/TransT/pytracking/networks/CrossViT_22k_ep0108.pth.tar'

    torch.set_num_threads(1)
    # torch.cuda.set_device(CUDA_ID)  # set GPU id
    save_root = '/home/v-xfei/Public/dataset/vot_tools/vot21/vis/'
    if VIS and (not os.path.exists(save_root)):
        os.mkdir(save_root)
    # Absolute path of the model
    print(net_path)
    # create model
    net = NetWithBackbone(net_path=net_path, use_gpu=True)
    tracker = TrackingNC_ALPHA(name='dualtfr160', net=net, window_penalty=0.49, exemplar_size=224,
                               instance_size=224,
                               refine_path='/home/v-xfei/Public/project/TransT/pysot_toolkit/ar.pth.tar',
                               search_z=3, search_x=3.0)

    handle = vot.VOT("mask")
    selection = handle.region()
    imagefile = handle.frame()

    if VIS:
        '''for vis'''
        seq_name = imagefile.split('/')[-3]
        save_v_dir = os.path.join(save_root,seq_name)
        if not os.path.exists(save_v_dir):
            os.mkdir(save_v_dir)
        cur_time = int(time.time() % 10000)
        save_dir = os.path.join(save_v_dir, str(cur_time))
        if not os.path.exists(save_dir):
            os.mkdir(save_dir)

    image = cv2.cvtColor(cv2.imread(imagefile), cv2.COLOR_BGR2RGB) # Right
    # mask given by the toolkit ends with the target (zero-padding to the right and down is needed)
    mask = make_full_size(selection, (image.shape[1], image.shape[0]))
    tracker.initialize(image, mask)

    while True:
        imagefile = handle.frame()
        if not imagefile:
            break
        image = cv2.cvtColor(cv2.imread(imagefile), cv2.COLOR_BGR2RGB)  # Right
        b1, m = tracker.track(image)
        handle.report(m)
        if VIS:
            '''Visualization'''
            image_show = image[:,:,::-1].copy() # RGB --> BGR
            image_show = image_show.astype(np.float32)

            image_show[:,:,0] += 127.0 * m
            image_show = image_show.clip(0,255).astype(np.uint8)
            cv2.rectangle(image_show,(int(b1[0]),int(b1[1])),
                          (int(b1[0]+b1[2]),int(b1[1]+b1[3])),(0,0,255),2)
            save_path = os.path.join(save_dir, imagefile.split('/')[-1])
            cv2.imwrite(save_path,image_show)

#run_vot_exp()