
from __future__ import absolute_import

import numpy as np
import math
import torchvision.transforms.functional as tvisf
import cv2
import torch
import torch.nn.functional as F
import time
from pysot_toolkit.trackers.vot20_utils import *
import pysot_toolkit.vot2020 as vot

from pytracking.refine_modules.refine_module_vot20 import RefineModule

import os
from pysot_toolkit.trackers.net_wrappers import NetWithBackbone
from pysot_toolkit.classifier.base_classifier import BaseClassifier
from pysot_toolkit.config_motion import cfg

from PIL import Image
import random

def iou(box1, box2, wh=False):
    """
    compute the iou of two boxes.
    Args:
        box1, box2: [xmin, ymin, xmax, ymax] (wh=False) or [xcenter, ycenter, w, h] (wh=True)
        wh: the format of coordinate.
    Return:
        iou: iou of box1 and box2.
    """
    if wh == False:
        xmin1, ymin1, xmax1, ymax1 = box1
        xmin2, ymin2, xmax2, ymax2 = box2
    else:
        xmin1, ymin1 = int(box1[0] - box1[2] / 2.0), int(box1[1] - box1[3] / 2.0)
        xmax1, ymax1 = int(box1[0] + box1[2] / 2.0), int(box1[1] + box1[3] / 2.0)
        xmin2, ymin2 = int(box2[0] - box2[2] / 2.0), int(box2[1] - box2[3] / 2.0)
        xmax2, ymax2 = int(box2[0] + box2[2] / 2.0), int(box2[1] + box2[3] / 2.0)

    xx1 = np.max([xmin1, xmin2])
    yy1 = np.max([ymin1, ymin2])
    xx2 = np.min([xmax1, xmax2])
    yy2 = np.min([ymax1, ymax2])

    area1 = (xmax1 - xmin1) * (ymax1 - ymin1)
    area2 = (xmax2 - xmin2) * (ymax2 - ymin2)

    inter_area = (np.max([0, xx2 - xx1])) * (np.max([0, yy2 - yy1]))
    iou = inter_area / (area1 + area2 - inter_area + 1e-6)

    return iou


class TrackingNC_motion(object):

    def __init__(self, name, net, window_penalty=0.49, exemplar_size=224, instance_size=224,
                 refine_path='/home/v-xfei/Public/project/TransT/pysot_toolkit/ar.pth.tar',  search_z=3, search_x=3):
        self.name = name
        self.net = net
        self.window_penalty = window_penalty
        self.exemplar_size = exemplar_size
        self.instance_size = instance_size
        self.net = net
      #  self.net.eval()

        self.alpha = RefineModule(refine_path, 2, 256)
        self.THRES = 0.65

        self.classifier = BaseClassifier(self.net)
        self.lost_count = 0

        self.search_z = search_z
        self.search_x = search_x


        #motion
        self.mem_step = cfg.TRACK.MEM_STEP
        self.mem_len = cfg.TRACK.MEM_LEN
        self.st_mem_coef = cfg.TRACK.ST_MEM_COEF
        self.mem_sink_idx = cfg.TRACK.MEM_SINK_IDX
        self.score_size = 14

    def _convert_score(self, score):

        score = score.permute(2, 1, 0).contiguous().view(2, -1).permute(1, 0)
        score = F.softmax(score, dim=1).data[:, 0].cpu().numpy()
        return score

    def _convert_bbox(self, delta):

        delta = delta.permute(2, 1, 0).contiguous().view(4, -1)
        delta = delta.data.cpu().numpy()

        return delta

    def _bbox_clip(self, cx, cy, width, height, boundary):
        cx = max(0, min(cx, boundary[1]))
        cy = max(0, min(cy, boundary[0]))
        width = max(10, min(width, boundary[1]))
        height = max(10, min(height, boundary[0]))
        return cx, cy, width, height

    def get_subwindow(self, im, pos, model_sz, original_sz, avg_chans):
        """
        args:
            im: rgb based image
            pos: center position
            model_sz: exemplar size
            original_sz: original size
            avg_chans: channel average
        """
        if isinstance(pos, float):
            pos = [pos, pos]
        sz = original_sz
        im_sz = im.shape
        c = (original_sz + 1) / 2
        # context_xmin = round(pos[0] - c) # py2 and py3 round
        context_xmin = np.floor(pos[0] - c + 0.5)
        context_xmax = context_xmin + sz - 1
        # context_ymin = round(pos[1] - c)
        context_ymin = np.floor(pos[1] - c + 0.5)
        context_ymax = context_ymin + sz - 1
        left_pad = int(max(0., -context_xmin))
        top_pad = int(max(0., -context_ymin))
        right_pad = int(max(0., context_xmax - im_sz[1] + 1))
        bottom_pad = int(max(0., context_ymax - im_sz[0] + 1))

        context_xmin = context_xmin + left_pad
        context_xmax = context_xmax + left_pad
        context_ymin = context_ymin + top_pad
        context_ymax = context_ymax + top_pad

        r, c, k = im.shape
        if any([top_pad, bottom_pad, left_pad, right_pad]):
            size = (r + top_pad + bottom_pad, c + left_pad + right_pad, k)
            te_im = np.zeros(size, np.uint8)
            te_im[top_pad:top_pad + r, left_pad:left_pad + c, :] = im
            if top_pad:
                te_im[0:top_pad, left_pad:left_pad + c, :] = avg_chans
            if bottom_pad:
                te_im[r + top_pad:, left_pad:left_pad + c, :] = avg_chans
            if left_pad:
                te_im[:, 0:left_pad, :] = avg_chans
            if right_pad:
                te_im[:, c + left_pad:, :] = avg_chans
            im_patch = te_im[int(context_ymin):int(context_ymax + 1),
                       int(context_xmin):int(context_xmax + 1), :]
        else:
            im_patch = im[int(context_ymin):int(context_ymax + 1),
                       int(context_xmin):int(context_xmax + 1), :]

        if not np.array_equal(model_sz, original_sz):
            im_patch = cv2.resize(im_patch, (model_sz, model_sz))
        im_patch = im_patch.transpose(2, 0, 1)
        im_patch = im_patch[np.newaxis, :, :, :]
        im_patch = im_patch.astype(np.float32)
        im_patch = torch.from_numpy(im_patch)
        im_patch = im_patch.cuda()
        return im_patch



    def initialize_features(self):
        if not getattr(self, 'features_initialized', False):
            self.net.initialize()
        self.features_initialized = True

    def initialize(self, image, mask):
        tic = time.time()
        hanning = np.hanning(14)  # 14
        window = np.outer(hanning, hanning)
        self.window = window.flatten()
        # Initialize
        self.initialize_features()
        """"""""""""
        if isinstance(mask, list):
            region = mask
        elif isinstance(mask, dict):
            bbox = mask['init_bbox']
            self.center_pos = np.array([bbox[0] + bbox[2] / 2,
                                        bbox[1] + bbox[3] / 2])
            self.size = np.array([bbox[2], bbox[3]])
            region = bbox

        else:
            region = rect_from_mask(mask)

        self.H, self.W, _ = image.shape
        gt_bbox_np = np.array(region).astype(np.float32)


        '''initilize refinement module for specific video'''
        self.alpha.initialize(image, gt_bbox_np)

        bbox = gt_bbox_np
        self.center_pos = np.array([bbox[0] + bbox[2] / 2,
                                    bbox[1] + bbox[3] / 2])
        self.size = np.array([bbox[2], bbox[3]])

        # calculate z crop size
        w_z = self.size[0] + (self.search_z - 1) * ((self.size[0] + self.size[1]) * 0.5)  #
        h_z = self.size[1] + (self.search_z - 1) * ((self.size[0] + self.size[1]) * 0.5)
        s_z = math.ceil(math.sqrt(w_z * h_z))

        scale_z = 224 / s_z


        # calculate channle average
        self.channel_average = np.mean(image, axis=(0, 1))

        # get crop
        z_crop = self.get_subwindow(image, self.center_pos,
                                    self.exemplar_size,
                                    s_z, self.channel_average)

        # normalize
        z_crop = z_crop.float().mul(1.0 / 255.0).clamp(0.0, 1.0)
        self.mean = [0.485, 0.456, 0.406]
        self.std = [0.229, 0.224, 0.225]
        self.inplace = False
        z_crop[0] = tvisf.normalize(z_crop[0], self.mean, self.std, self.inplace)

        # initialize template feature
        self.net.template(z_crop)
        out = {'time': time.time() - tic}

        ###online###
        s_xx_online = s_z *(224 / 128)
        x_crop_online = self.get_subwindow(image, self.center_pos,
                                    224,
                                    s_xx_online, self.channel_average)

        self.classifier.initialize(x_crop_online.type(torch.FloatTensor), bbox)


        # rpt process
        self.net.template(z_crop)
        self.features = [self.net.zf] * self.mem_len

        ##Next part for doing some strategy about motion
        self.frame_num = 0
        self.dist = []
        self.dist_x = []
        self.dist_y = []
        self.speed = 0.0
        self.speed_x = 0.0
        self.speed_y = 0.0
        self.disturbance_close_to_target = False
        self.disturbance_away_from_target = False
        self.disturbance_in_target = False
        self.previous_t_d_distance = 0
        self.inside_target_pos_x = 0
        self.inside_target_pos_y = 0

        return out

    def post_processing(self, score, pred_bbox, scale_z=None):

        pscore = score * (1 - cfg.TRACK.WINDOW_INFLUENCE) + self.window * cfg.TRACK.WINDOW_INFLUENCE
        best_idx = np.argmax(pscore)

        final_lr_score = score[best_idx]
        final_bbox = pred_bbox[:,best_idx]
        final_score = pscore[best_idx] * self.filter_confidence
        lr = cfg.TRACK.LR * final_lr_score
        # print(str(final_score))
        if final_lr_score > 0.40:
            lr = max(lr, 0.50)
        box = final_bbox

        return box, final_score, lr, final_lr_score

    def motion_strategy(self, score, pred_bbox, scale_z):
        pscore = score.copy()
        score = score.reshape(self.score_size, self.score_size)

        inside_weight = np.zeros_like(score)
        inside_width = cfg.TRACK.SCORE_INSIDE_WIDTH
        inside_right_idx = int(self.score_size - (self.score_size - inside_width) / 2)
        inside_left_idx = int(self.score_size - (self.score_size - inside_width) / 2 - inside_width)
        inside_weight[inside_left_idx:inside_right_idx, inside_left_idx:inside_right_idx] = np.ones(
            (inside_width, inside_width))

        outside_width = cfg.TRACK.SCORE_OUTSIDE_WIDTH
        outside_right_idx = int(self.score_size - (self.score_size - outside_width) / 2)
        outside_left_idx = int(self.score_size - (self.score_size - outside_width) / 2 - outside_width)

        outside_weight = np.zeros_like(score)
        outside_weight[outside_left_idx:outside_right_idx, outside_left_idx:outside_right_idx] = np.ones(
            (outside_width, outside_width))
        outside_weight = outside_weight - inside_weight

        inside_score = score * inside_weight
        outside_score = score * outside_weight

        if outside_score.max() > 0.3 and inside_score.max() > 0.4:
            inside_score = inside_score.reshape(-1)
            outside_score = outside_score.reshape(-1)
            inside_box, final_score, lr, final_lr_score = self.post_processing(inside_score, pred_bbox, scale_z)
            inside_pos_x = self.center_pos[0] + inside_box[0] / scale_z
            inside_pos_y = self.center_pos[1] + inside_box[1] / scale_z

            outside_box, final_score, lr, final_lr_score = self.post_processing(outside_score, pred_bbox, scale_z)
            outside_pos_x = self.center_pos[0] + outside_box[0] / scale_z
            outside_pos_y = self.center_pos[1] + outside_box[1] / scale_z

            target_disturbance_distance = np.sqrt(
                (outside_pos_x - inside_pos_x) ** 2 + (outside_pos_y - inside_pos_y) ** 2)

            if self.previous_t_d_distance == 0:
                self.previous_t_d_distance = target_disturbance_distance
            else:
                if target_disturbance_distance - self.previous_t_d_distance < 0:
                    self.disturbance_close_to_target = True

                    self.inside_target_pos_x = inside_pos_x
                    self.inside_target_pos_y = inside_pos_y
                    self.t_d_reset_count = 0
                elif target_disturbance_distance - self.previous_t_d_distance > 0 and self.disturbance_in_target is True:
                    self.disturbance_away_from_target = True

            box = target_box = inside_box
        else:
            box, final_score, lr, final_lr_score = self.post_processing(pscore, pred_bbox, scale_z)
            if self.disturbance_close_to_target is True:
                self.disturbance_in_target = True
                self.previous_t_d_distance = 0
                inside_box = box
                inside_pos_x = self.center_pos[0] + inside_box[0] / scale_z
                inside_pos_y = self.center_pos[1] + inside_box[1] / scale_z
                self.t_d_reset_count = self.t_d_reset_count + 1
                if self.t_d_reset_count == 10:
                    self.disturbance_close_to_target = False
                    self.disturbance_in_target = False
                    self.disturbance_away_from_target = False
            inside_box = box
            outside_box = box

        if self.disturbance_away_from_target is True:
            inside_pos_x = self.center_pos[0] + inside_box[0] / scale_z
            inside_pos_y = self.center_pos[1] + inside_box[1] / scale_z
            target_inside_distance = np.sqrt(
                (self.inside_target_pos_x - inside_pos_x) ** 2 + (self.inside_target_pos_y - inside_pos_y) ** 2)
            outside_pos_x = self.center_pos[0] + outside_box[0] / scale_z
            outside_pos_y = self.center_pos[1] + outside_box[1] / scale_z
            target_outside_distance = np.sqrt(
                (outside_pos_x - self.inside_target_pos_x) ** 2 + (outside_pos_y - self.inside_target_pos_y) ** 2)
            if target_inside_distance > target_outside_distance:
                disturbance_box = inside_box
                target_box = outside_box
            else:
                disturbance_box = outside_box
                target_box = inside_box

            self.disturbance_close_to_target = False
            self.disturbance_in_target = False
            self.disturbance_away_from_target = False
            box = target_box

        return box, final_score, lr, final_lr_score

    def track(self, image, info=None):
        self.temp_max = 0
        self.frame_num = self.frame_num+1
        # calculate x crop size
        w_x = self.size[0] + (self.search_x - 1) * ((self.size[0] + self.size[1]) * 0.5)
        h_x = self.size[1] + (self.search_x - 1) * ((self.size[0] + self.size[1]) * 0.5)
        self.s_x = math.ceil(math.sqrt(w_x * h_x))
        scale_z = 224 / self.s_x

        # get crop
        x_crop = self.get_subwindow(image, self.center_pos,
                                    self.instance_size,
                                    round(self.s_x), self.channel_average)

        # normalize
        x_crop = x_crop.float().mul(1.0 / 255.0).clamp(0.0, 1.0)
        x_crop[0] = tvisf.normalize(x_crop[0], self.mean, self.std, self.inplace)

        # rpt process
        pred_bbox = []
        scores_transnc = []

        for idx in range(self.mem_len):
            with torch.no_grad():
                if idx == 0:
                    outputs = self.net.track_motion(x_crop, self.features[idx])
                    filter_search_feature = outputs['feature']
                else:
                    outputs = self.net.tracking_motion(outputs['search_feat'], self.features[idx])
                scores_transnc.append(outputs['score'])
                pred_bbox.append(outputs['bbox'])

        if self.mem_len > 1:
            fuse_func = lambda x: x[0] * self.st_mem_coef + np.stack(x[1:], axis=0).mean(axis=0) * (
                        1 - self.st_mem_coef)
            scores_transnc = fuse_func(scores_transnc)
        else:
            scores_transnc = scores_transnc[0]
        pred_bbox = pred_bbox[0]  ##using the 1st template to predict boundingbox


        ### online
        self.classifier.model.instance_f(filter_search_feature)
        flag, scores_match = self.classifier.track()

        scores_match = Image.fromarray(scores_match.detach().cpu().numpy())
        scores_match = np.array(scores_match.resize((14, 14))).flatten()
        scores_match = np.clip(scores_match, 0, 0.999)

        def change(r):
            return np.maximum(r, 1. / r)

        def sz(w, h):
            pad = (w + h) * 0.5
            return np.sqrt((w + pad) * (h + pad))


        pscore = scores_transnc
       # print(str(flag))
       # print(str(np.max(scores_match)))
        self.filter_confidence = np.max(scores_match)

        pscore = pscore * (
                    1 - cfg.TRACK.ONLINE_CLASSIFICATION_INFLUENCE) + scores_match * cfg.TRACK.ONLINE_CLASSIFICATION_INFLUENCE

        if self.speed_x > cfg.TRACK.SPEED_INFLUENCE * self.size[0] or self.speed_y > cfg.TRACK.SPEED_INFLUENCE * \
                self.size[1] or self.speed > cfg.TRACK.SPEED_INFLUENCE * max(self.size):
            cfg.TRACK.WINDOW_INFLUENCE = cfg.TRACK.WINDOW_INFLUENCE_FAST
            self.search_x = cfg.TRACK.SEARCH_AREA_FAST
            cfg.TRACK.ONLINE_CLASSIFICATION_INFLUENCE = cfg.TRACK.ONLINE_CLASSIFICATION_INFLUENCE_FAST#0.6
            #print('fast')
        elif self.speed_x > self.size[0] or self.speed_y > self.size[1] or self.speed > max(self.size):
            cfg.TRACK.WINDOW_INFLUENCE = cfg.TRACK.WINDOW_INFLUENCE_MEDIUM
            self.search_x = cfg.TRACK.SEARCH_AREA_MEDIUM
            cfg.TRACK.ONLINE_CLASSIFICATION_INFLUENCE = cfg.TRACK.ONLINE_CLASSIFICATION_INFLUENCE_MEDIUM#0.4
            #print('medium')
        else:
            cfg.TRACK.WINDOW_INFLUENCE = cfg.TRACK.WINDOW_INFLUENCE_SLOW
            self.search_x = cfg.TRACK.SEARCH_AREA_SLOW
            cfg.TRACK.ONLINE_CLASSIFICATION_INFLUENCE = cfg.TRACK.ONLINE_CLASSIFICATION_INFLUENCE_SLOW #0.2
            #print('slow')

       # if self.filter_confidence < 0.2:
        #cfg.TRACK.WINDOW_INFLUENCE = 0.49
       # self.search_x = 3
        #print('usure')


        box, best_score, lr, final_lr_score = self.motion_strategy(pscore.copy(), pred_bbox, scale_z)

        #final_bbox = box
        #bbox = box / scale_z

        bbox = box * self.s_x
        cx = bbox[0] + self.center_pos[0] - self.s_x / 2
        cy = bbox[1] + self.center_pos[1] - self.s_x / 2
        width = bbox[2]
        height = bbox[3]



        #cx = self.center_pos[0] + bbox[0]
        #cy = self.center_pos[1] + bbox[1]

        self.dist.append(math.sqrt(bbox[0] ** 2 + bbox[1] ** 2))
        self.dist_x.append(np.abs(bbox[0]))
        self.dist_y.append(np.abs(bbox[1]))

        if len(self.dist) < cfg.TRACK.SPEED_LAST_CALC:
            self.speed = max(self.dist)
            self.speed_x = max(self.dist_x)
            self.speed_y = max(self.dist_y)
        else:
            self.speed = max(self.dist[-cfg.TRACK.SPEED_LAST_CALC:])
            self.speed_x = max(self.dist_x[-cfg.TRACK.SPEED_LAST_CALC:])
            self.speed_y = max(self.dist_y[-cfg.TRACK.SPEED_LAST_CALC:])

        # smooth bbox
        #print(str(lr))
        width = self.size[0] * (1 - lr) + bbox[2] * lr
        height = self.size[1] * (1 - lr) + bbox[3] * lr

        # clip boundary
        cx, cy, width, height = self._bbox_clip(cx, cy, width,
                                                height, image.shape[:2])

        # udpate state
        self.center_pos = np.array([cx, cy])
        self.size = np.array([width, height])

        bbox = [cx - width / 2,
                cy - height / 2,
                width,
                height]

        ##Next part for updating Online Classification and SiameseReppointsTracker template
        #if (final_lr_score > 0.3):
        self.classifier.update(bbox, flag)

        if (self.frame_num - 1) % self.mem_step == 0 and self.frame_num > 1 and final_lr_score > 0.45 and self.mem_len > 1:
            # calculate z crop size
            w_z = self.size[0] + (3 - 1) * ((self.size[0] + self.size[1]) * 0.5)  #
            h_z = self.size[1] + (3 - 1) * ((self.size[0] + self.size[1]) * 0.5)
            s_z = math.ceil(math.sqrt(w_z * h_z))

            self.channel_average = np.mean(image, axis=(0, 1))
            self.z_crop = self.get_subwindow(image, self.center_pos, self.exemplar_size, s_z,
                                                 self.channel_average)
            with torch.no_grad():
                #self.net.template_short_term(self.z_crop)
                feat = self.net.get_template(self.z_crop)
                #print(str(feat.shape))
            self.features.pop(self.mem_sink_idx)
            self.features.append(feat)

####################################################################################################
        out = {'target_bbox': bbox,
               'best_score': pscore}

        pred_bbox = out['target_bbox']
        '''Step1: Post-Process'''
        #x1, y1, w, h = pred_bbox
        # add boundary and min size limit
        #x1, y1, x2, y2 = bbox_clip(x1, y1, x1 + w, y1 + h, (self.H, self.W))
        #w = x2 - x1
        #h = y2 - y1
        #bbox_new = [x1, y1, w, h]
        '''Step2: Mask report'''
        pred_mask = self.alpha.refine(image, np.array(pred_bbox))
        final_mask = (pred_mask > self.THRES).astype(np.uint8)

        return pred_bbox, final_mask  # out #







def run_vot_exp(VIS=False):

    torch.set_num_threads(1)

    def set_seed(seed):
        random.seed(seed)
        os.environ['PYTHONHASHSEED'] = str(seed)
        np.random.seed(seed)
        torch.manual_seed(seed)
        torch.cuda.manual_seed(seed)
        torch.backends.cudnn.benchmark = True
        torch.backends.cudnn.deterministic = True
        torch.backends.cudnn.enabled = True
    seed_test = 1000
    set_seed(seed_test)

    # torch.cuda.set_device(CUDA_ID)  # set GPU id
    save_root = '/home/v-xfei/Public/dataset/vot_tools/vot21/vis/'
    if VIS and (not os.path.exists(save_root)):
        os.mkdir(save_root)
    net_path = '/home/v-xfei/Public/project/TransT/pytracking/networks/CrossViT_ep0099.pth.tar'  # Absolute path of the model
    print(net_path)
    # create model
    net = NetWithBackbone(net_path=net_path, use_gpu=True)
    tracker = TrackingNC_motion(name='dualtfr160', net=net, window_penalty=0.49, exemplar_size=224, instance_size=224,
                                refine_path='/tmp/code/pysot_toolkit/ar.pth.tar')

    handle = vot.VOT("mask")
    selection = handle.region()
    imagefile = handle.frame()

    if VIS:
        '''for vis'''
        seq_name = imagefile.split('/')[-3]
        save_v_dir = os.path.join(save_root,seq_name)
        if not os.path.exists(save_v_dir):
            os.mkdir(save_v_dir)
        cur_time = int(time.time() % 10000)
        save_dir = os.path.join(save_v_dir, str(cur_time))
        if not os.path.exists(save_dir):
            os.mkdir(save_dir)

    image = cv2.cvtColor(cv2.imread(imagefile), cv2.COLOR_BGR2RGB) # Right
    # mask given by the toolkit ends with the target (zero-padding to the right and down is needed)
    mask = make_full_size(selection, (image.shape[1], image.shape[0]))
    tracker.initialize(image, mask)

    while True:
        imagefile = handle.frame()
        if not imagefile:
            break
        image = cv2.cvtColor(cv2.imread(imagefile), cv2.COLOR_BGR2RGB)  # Right
        b1, m = tracker.track(image)
        handle.report(m)
        if VIS:
            '''Visualization'''
            image_show = image[:,:,::-1].copy() # RGB --> BGR
            image_show = image_show.astype(np.float32)

            image_show[:,:,0] += 127.0 * m
            image_show = image_show.clip(0,255).astype(np.uint8)
            cv2.rectangle(image_show,(int(b1[0]),int(b1[1])),
                          (int(b1[0]+b1[2]),int(b1[1]+b1[3])),(0,0,255),2)
            save_path = os.path.join(save_dir, imagefile.split('/')[-1])
            cv2.imwrite(save_path,image_show)

#run_vot_exp()