
from __future__ import absolute_import

import numpy as np
import math
import torchvision.transforms.functional as tvisf
import cv2
import torch
import torch.nn.functional as F
import time
from pysot_toolkit.trackers.vot20_utils import *
import pysot_toolkit.vot2020 as vot

from pytracking.refine_modules.refine_module_vot20 import RefineModule

import os
from pysot_toolkit.trackers.net_wrappers import NetWithBackbone
from pysot_toolkit.classifier.base_classifier import BaseClassifier
from pysot_toolkit.classifier.config import cfg
from PIL import Image
import random

def iou(box1, box2, wh=False):
    """
    compute the iou of two boxes.
    Args:
        box1, box2: [xmin, ymin, xmax, ymax] (wh=False) or [xcenter, ycenter, w, h] (wh=True)
        wh: the format of coordinate.
    Return:
        iou: iou of box1 and box2.
    """
    if wh == False:
        xmin1, ymin1, xmax1, ymax1 = box1
        xmin2, ymin2, xmax2, ymax2 = box2
    else:
        xmin1, ymin1 = int(box1[0] - box1[2] / 2.0), int(box1[1] - box1[3] / 2.0)
        xmax1, ymax1 = int(box1[0] + box1[2] / 2.0), int(box1[1] + box1[3] / 2.0)
        xmin2, ymin2 = int(box2[0] - box2[2] / 2.0), int(box2[1] - box2[3] / 2.0)
        xmax2, ymax2 = int(box2[0] + box2[2] / 2.0), int(box2[1] + box2[3] / 2.0)

    xx1 = np.max([xmin1, xmin2])
    yy1 = np.max([ymin1, ymin2])
    xx2 = np.min([xmax1, xmax2])
    yy2 = np.min([ymax1, ymax2])

    area1 = (xmax1 - xmin1) * (ymax1 - ymin1)
    area2 = (xmax2 - xmin2) * (ymax2 - ymin2)

    inter_area = (np.max([0, xx2 - xx1])) * (np.max([0, yy2 - yy1]))
    iou = inter_area / (area1 + area2 - inter_area + 1e-6)

    return iou


class TrackingNC_ALPHA_ON(object):

    def __init__(self, name, net, window_penalty=0.49,
                 sz_penalty = 0.5, thres=0.65, avg_motion_thes =0.20, avg_conf_thes=0.90, window_penalty_p=0.30,
                 search_mask = 2.0,
                 update_num = 20,  cls_uc = 0.6, cls_c= 0.2,
                 larger_area = 4.0,

                 refine_path='/tmp/code/pysot_toolkit/ar.pth.tar'):
        self.name = name
        self.net = net
        self.window_penalty = window_penalty
        self.exemplar_size = 224
        self.instance_size = 224
        self.net = self.net

        self.alpha = RefineModule(refine_path, 2, 256)
        self.THRES = thres

        self.classifier = BaseClassifier(self.net)
        self.lost_count = 0

        self.search_z = 3.0
        self.search_x = 3.0
        self.large_area = larger_area

        self.window_penalty_n = window_penalty
        self.window_penalty = window_penalty

        self.penalty_k = sz_penalty
        self.avg_motion_thes = avg_motion_thes
        self.avg_conf_thes = avg_conf_thes
        self.window_penalty_p = window_penalty_p

        self.motion_h = []
        self.motion_ref = []
        self.confidence_h = []

        self.search_mask = search_mask
        self.COEE_CLASS = 0.3

        self.update_num = update_num
        self.COEE_CLASS_uncertain = cls_uc
        self.COEE_CLASS_certain = cls_c


    def _convert_score(self, score):

        score = score.permute(2, 1, 0).contiguous().view(2, -1).permute(1, 0)
        score = F.softmax(score, dim=1).data[:, 0].cpu().numpy()
        return score

    def _convert_bbox(self, delta):

        delta = delta.permute(2, 1, 0).contiguous().view(4, -1)
        delta = delta.data.cpu().numpy()

        return delta

    def _bbox_clip(self, cx, cy, width, height, boundary):
        cx = max(0, min(cx, boundary[1]))
        cy = max(0, min(cy, boundary[0]))
        width = max(10, min(width, boundary[1]))
        height = max(10, min(height, boundary[0]))
        return cx, cy, width, height

    def get_subwindow(self, im, pos, model_sz, original_sz, avg_chans):
        """
        args:
            im: rgb based image
            pos: center position
            model_sz: exemplar size
            original_sz: original size
            avg_chans: channel average
        """
        if isinstance(pos, float):
            pos = [pos, pos]
        sz = original_sz
        im_sz = im.shape
        c = (original_sz + 1) / 2
        # context_xmin = round(pos[0] - c) # py2 and py3 round
        context_xmin = np.floor(pos[0] - c + 0.5)
        context_xmax = context_xmin + sz - 1
        # context_ymin = round(pos[1] - c)
        context_ymin = np.floor(pos[1] - c + 0.5)
        context_ymax = context_ymin + sz - 1
        left_pad = int(max(0., -context_xmin))
        top_pad = int(max(0., -context_ymin))
        right_pad = int(max(0., context_xmax - im_sz[1] + 1))
        bottom_pad = int(max(0., context_ymax - im_sz[0] + 1))

        context_xmin = context_xmin + left_pad
        context_xmax = context_xmax + left_pad
        context_ymin = context_ymin + top_pad
        context_ymax = context_ymax + top_pad

        r, c, k = im.shape
        if any([top_pad, bottom_pad, left_pad, right_pad]):
            size = (r + top_pad + bottom_pad, c + left_pad + right_pad, k)
            te_im = np.zeros(size, np.uint8)
            te_im[top_pad:top_pad + r, left_pad:left_pad + c, :] = im
            if top_pad:
                te_im[0:top_pad, left_pad:left_pad + c, :] = avg_chans
            if bottom_pad:
                te_im[r + top_pad:, left_pad:left_pad + c, :] = avg_chans
            if left_pad:
                te_im[:, 0:left_pad, :] = avg_chans
            if right_pad:
                te_im[:, c + left_pad:, :] = avg_chans
            im_patch = te_im[int(context_ymin):int(context_ymax + 1),
                       int(context_xmin):int(context_xmax + 1), :]
        else:
            im_patch = im[int(context_ymin):int(context_ymax + 1),
                       int(context_xmin):int(context_xmax + 1), :]

        if not np.array_equal(model_sz, original_sz):
            im_patch = cv2.resize(im_patch, (model_sz, model_sz))
        im_patch = im_patch.transpose(2, 0, 1)
        im_patch = im_patch[np.newaxis, :, :, :]
        im_patch = im_patch.astype(np.float32)
        im_patch = torch.from_numpy(im_patch)
        im_patch = im_patch.cuda()
        return im_patch


    def initialize_features(self):
        if not getattr(self, 'features_initialized', False):
            self.net.initialize()
        self.features_initialized = True

    def initialize(self, image, mask):
        tic = time.time()
        hanning = np.hanning(14)  # 14
        window = np.outer(hanning, hanning)
        self.window = window.flatten()
        # Initialize
        self.initialize_features()
        """"""""""""
        if isinstance(mask, list):
            region = mask
        elif isinstance(mask, dict):
            bbox = mask['init_bbox']
            self.center_pos = np.array([bbox[0] + bbox[2] / 2,
                                        bbox[1] + bbox[3] / 2])
            self.size = np.array([bbox[2], bbox[3]])
            region = bbox

        else:
            region = rect_from_mask(mask)

        self.H, self.W, _ = image.shape
        gt_bbox_np = np.array(region).astype(np.float32)


        '''initilize refinement module for specific video'''
        self.alpha.initialize(image, gt_bbox_np)

        bbox = gt_bbox_np
        self.center_pos = np.array([bbox[0] + bbox[2] / 2,
                                    bbox[1] + bbox[3] / 2])
        self.size = np.array([bbox[2], bbox[3]])

        # calculate z crop size
        w_z = self.size[0] + (self.search_z - 1) * ((self.size[0] + self.size[1]) * 0.5)  #
        h_z = self.size[1] + (self.search_z - 1) * ((self.size[0] + self.size[1]) * 0.5)
        s_z = math.ceil(math.sqrt(w_z * h_z))


        # calculate channle average
        self.channel_average = np.mean(image, axis=(0, 1))

        # get crop
        z_crop = self.get_subwindow(image, self.center_pos,
                                    self.exemplar_size,
                                    s_z, self.channel_average)

        # normalize
        z_crop = z_crop.float().mul(1.0 / 255.0).clamp(0.0, 1.0)
        self.mean = [0.485, 0.456, 0.406]
        self.std = [0.229, 0.224, 0.225]
        self.inplace = False
        z_crop[0] = tvisf.normalize(z_crop[0], self.mean, self.std, self.inplace)

        # initialize template feature
        self.net.template(z_crop)
        out = {'time': time.time() - tic}

        ###online###

        s_xx_online = s_z *(224 / 128)
        x_crop_online = self.get_subwindow(image, self.center_pos,
                                    224,
                                    s_xx_online, self.channel_average)

        self.classifier.initialize(x_crop_online.type(torch.FloatTensor), bbox)
        if cfg.TRACK.TEMPLATE_UPDATE:
            with torch.no_grad():
                self.net.template_short_term(z_crop)
        self.frame_num =1
        return out

    def track(self, image, info=None):
        self.temp_max = 0
        self.frame_num = self.frame_num+1
        # calculate x crop size
        w_x = self.size[0] + (self.search_x - 1) * ((self.size[0] + self.size[1]) * 0.5)
        h_x = self.size[1] + (self.search_x - 1) * ((self.size[0] + self.size[1]) * 0.5)
        s_x = math.ceil(math.sqrt(w_x * h_x))

        # get crop
        x_crop = self.get_subwindow(image, self.center_pos,
                                    self.instance_size,
                                    round(s_x), self.channel_average)

        # normalize
        x_crop = x_crop.float().mul(1.0 / 255.0).clamp(0.0, 1.0)
        x_crop[0] = tvisf.normalize(x_crop[0], self.mean, self.std, self.inplace)

        # track
        outputs = self.net.track(x_crop)

        score = self._convert_score(outputs['pred_logits'])
        pred_bbox = self._convert_bbox(outputs['pred_boxes'])


        ### online
        self.classifier.model.instance(x_crop)
        flag, s = self.classifier.track()
        if flag == 'not_found':
            self.lost_count += 1
        else:
            self.lost_count = 0

        def normalize(score):
            score = (score - np.min(score)) / (np.max(score) - np.min(score))
            return score

        confidence = Image.fromarray(s.detach().cpu().numpy())
        confidence = np.array(confidence.resize((14, 14))).flatten()
        pscore = score * (1 - self.COEE_CLASS) + \
                 normalize(confidence) * self.COEE_CLASS


        ####################
        # window penalty
        pscore = pscore * (1 - self.window_penalty) + \
                 self.window * self.window_penalty

        best_idx = np.argmax(pscore)
        bbox = pred_bbox[:, best_idx]
        bbox = bbox * s_x
        cx = bbox[0] + self.center_pos[0] - s_x / 2
        cy = bbox[1] + self.center_pos[1] - s_x / 2
        width = bbox[2]
        height = bbox[3]

        best_idx = np.argmax(pscore)
        best_score = np.max(pscore)

        motion = ((cx - self.center_pos[0]) ** 2 + (cy - self.center_pos[1]) ** 2) ** 0.5
        self.motion_h.append(motion)
        self.motion_ref.append(s_x)
        self.confidence_h.append(best_score)

        if len(self.motion_h) > 3:
            avg_motion = np.mean(self.motion_h) / np.mean(self.motion_ref)
            avg_confidence = np.mean(self.confidence_h)
            if avg_motion > self.avg_motion_thes and avg_confidence < self.avg_conf_thes:  # 0.20 0.90
                self.search_x = self.large_area
                self.window_penalty = self.window_penalty_p
                self.COEE_CLASS = self.COEE_CLASS_uncertain
            else:
                self.search_x = 3
                self.window_penalty = self.window_penalty_n
                self.COEE_CLASS = self.COEE_CLASS_certain
            if len(self.motion_h) == 10:
                del self.motion_h[0]
                del self.motion_ref[0]
                del self.confidence_h[0]
        else:
            self.search_x = 3

        if (self.frame_num % 1) == 0:
            self.net.context_update()

        if cfg.TRACK.TEMPLATE_UPDATE:
            outputs_st = self.net.track_st(x_crop)
            score_st = self._convert_score(outputs_st['pred_logits'])
            pred_bbox_st = self._convert_bbox(outputs_st['pred_boxes'])
            pscore_st = score_st * (1 - self.COEE_CLASS) + normalize(confidence) * self.COEE_CLASS
            pscore_st = pscore_st * (1 - self.window_penalty) + \
                     self.window * self.window_penalty

            best_idx_st = np.argmax(pscore_st)
            bbox_st = pred_bbox_st[:, best_idx]
            bbox_st = bbox_st * s_x
            cx_st = bbox_st[0] + self.center_pos[0] - s_x / 2
            cy_st = bbox_st[1] + self.center_pos[1] - s_x / 2
            width_st = bbox_st[2]
            height_st = bbox_st[3]
            if iou((cx_st, cy_st, width_st, height_st), (cx, cy, width, height), wh=True) >= cfg.TRACK.TAU_REGRESSION \
                and score_st[best_idx_st] - score[best_idx] >= cfg.TRACK.TAU_CLASSIFICATION:
                cx, cy, width, height, score, best_idx = cx_st, cy_st, width_st, height_st, score_st, best_idx_st

        # clip boundary
        cx, cy, width, height = self._bbox_clip(cx, cy, width,
                                                height, image.shape[:2])

        # update state
        self.center_pos = np.array([cx, cy])
        self.size = np.array([width, height])

        bbox = [cx - width / 2,
                cy - height / 2,
                width,
                height]

        out = {'target_bbox': bbox,
               'best_score': pscore}
        ###############
        self.classifier.update(bbox, flag)
        if (self.frame_num % self.update_num) == 0 and cfg.TRACK.TEMPLATE_UPDATE:
            if torch.max(s).item() >= cfg.TRACK.TARGET_UPDATE_THRESHOLD and flag != 'hard_negative':
                # calculate z crop size
                w_z = self.size[0] + (3 - 1) * ((self.size[0] + self.size[1]) * 0.5)  #
                h_z = self.size[1] + (3 - 1) * ((self.size[0] + self.size[1]) * 0.5)
                s_z = math.ceil(math.sqrt(w_z * h_z))

                if torch.max(s).item() > self.temp_max:
                    self.temp_max = torch.max(s).item()
                    self.channel_average = np.mean(image, axis=(0, 1))
                    self.z_crop = self.get_subwindow(image, self.center_pos, self.exemplar_size, s_z,
                                                     self.channel_average)

                if (self.frame_num - 1) % cfg.TRACK.TARGET_UPDATE_SKIPPING == 0:
                    self.temp_max = 0
                    with torch.no_grad():
                        self.net.template_short_term(self.z_crop)


        pred_bbox = out['target_bbox']
        '''Step1: Post-Process'''
        x1, y1, w, h = pred_bbox
        # add boundary and min size limit
        x1, y1, x2, y2 = bbox_clip(x1, y1, x1 + w, y1 + h, (self.H, self.W))
        w = x2 - x1
        h = y2 - y1

        bbox_new = [x1, y1, w, h]
        '''Step2: Mask report'''
        pred_mask = self.alpha.refine(image, np.array(bbox_new))
        final_mask = (pred_mask > self.THRES).astype(np.uint8)

        return out#, final_mask #bbox_new, final_mask  #

def set_seed(seed):
    random.seed(seed)
    os.environ['PYTHONHASHSEED'] = str(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.backends.cudnn.benchmark = True
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.enabled = True

def run_vot_exp(VIS=False):
    # index = int(pth) + 350
    pth_path = '/tmp/code/pytracking/networks/CrossViT_ep0098.pth.tar'
    #########################################
    import json
    save_path = '/tmp/code/pt_dirs1/hp_rand.json'
    with open(save_path, encoding='utf8') as fp:
        hp = json.load(fp)
    wp = hp['window_penalty']
    sp = hp['sz_penalty']
    ts = hp['threshold']
    amt = hp['avg_motion_thes']
    act = hp['avg_conf_thes']
    wpp = hp['window_penalty_p']
    sea_mask = hp['search_mask']
    net_path = pth_path

    up_num = hp['update_num']# = 20
    cls_uc = hp['cls_uc']#0.6
    cls_c = hp['cls_c'] #0.2
    larger_area =  hp['larger_area']#4.0

    print(net_path)
    seed_test = hp['seed_test']
    set_seed(seed_test)

    torch.set_num_threads(1)

    print(net_path)
    # create model
    net = NetWithBackbone(net_path=net_path, use_gpu=True)
    tracker = TrackingNC_ALPHA_ON(net=net,
                                  window_penalty=wp,sz_penalty=sp, thres=ts,
                                  avg_motion_thes=amt,avg_conf_thes=act,
                                  window_penalty_p=wpp,search_mask = sea_mask,

                                  update_num=up_num, cls_uc=cls_uc, cls_c=cls_c,
                                  larger_area=larger_area,

                                  refine_path = '/tmp/code/pysot_toolkit/ar.pth.tar',
                                  )

    handle = vot.VOT("mask")
    selection = handle.region()
    imagefile = handle.frame()

    image = cv2.cvtColor(cv2.imread(imagefile), cv2.COLOR_BGR2RGB) # Right
    # mask given by the toolkit ends with the target (zero-padding to the right and down is needed)
    mask = make_full_size(selection, (image.shape[1], image.shape[0]))
    tracker.initialize(image, mask)

    while True:
        imagefile = handle.frame()
        if not imagefile:
            break
        image = cv2.cvtColor(cv2.imread(imagefile), cv2.COLOR_BGR2RGB)  # Right
        b1, m = tracker.track(image)
        handle.report(m)

