
from __future__ import absolute_import

import numpy as np
import math
import torchvision.transforms.functional as tvisf
import cv2
import torch
import torch.nn.functional as F
import time
from pysot_toolkit.trackers.vot20_utils import *
import pysot_toolkit.vot2020 as vot

import os
from pysot_toolkit.trackers.net_wrappers import NetWithBackbone


class Tracker_blockT(object):

    def __init__(self, name, net_path,type='base', window_penalty=0.49, window_size=16, exemplar_size=128,
                 instance_size=256,cross_dict=None, net =None,
                 ):
        self.name = name
        self.net_path = net_path
        self.exemplar_size = exemplar_size
        self.instance_size = instance_size
        self.net = None
        self.window_size = int(window_size)
        self.type= type

        self.THRES = 0.65
        self.window_penalty = window_penalty

        self.penalty_k = 0.04
        self.test_lr = 0.95
        self.window_influence = 0.21
        self.cross_dict = cross_dict
        self.net = net


    def _convert_score(self, score):

        score = score.permute(2, 1, 0).contiguous().view(2, -1).permute(1, 0)
        score = F.softmax(score, dim=1).data[:, 0].cpu().numpy()
        return score

    def _convert_bbox(self, delta):

        delta = delta.permute(2, 1, 0).contiguous().view(4, -1)
        delta = delta.data.cpu().numpy()

        return delta

    def _bbox_clip(self, cx, cy, width, height, boundary):
        cx = max(0, min(cx, boundary[1]))
        cy = max(0, min(cy, boundary[0]))
        width = max(10, min(width, boundary[1]))
        height = max(10, min(height, boundary[0]))
        return cx, cy, width, height

    def get_subwindow(self, im, pos, model_sz, original_sz, avg_chans):
        """
        args:
            im: rgb based image
            pos: center position
            model_sz: exemplar size
            original_sz: original size
            avg_chans: channel average
        """
        if isinstance(pos, float):
            pos = [pos, pos]
        sz = original_sz
        im_sz = im.shape
        c = (original_sz + 1) / 2
        # context_xmin = round(pos[0] - c) # py2 and py3 round
        context_xmin = np.floor(pos[0] - c + 0.5)
        context_xmax = context_xmin + sz - 1
        # context_ymin = round(pos[1] - c)
        context_ymin = np.floor(pos[1] - c + 0.5)
        context_ymax = context_ymin + sz - 1
        left_pad = int(max(0., -context_xmin))
        top_pad = int(max(0., -context_ymin))
        right_pad = int(max(0., context_xmax - im_sz[1] + 1))
        bottom_pad = int(max(0., context_ymax - im_sz[0] + 1))

        context_xmin = context_xmin + left_pad
        context_xmax = context_xmax + left_pad
        context_ymin = context_ymin + top_pad
        context_ymax = context_ymax + top_pad

        r, c, k = im.shape
        if any([top_pad, bottom_pad, left_pad, right_pad]):
            size = (r + top_pad + bottom_pad, c + left_pad + right_pad, k)
            te_im = np.zeros(size, np.uint8)
            te_im[top_pad:top_pad + r, left_pad:left_pad + c, :] = im
            if top_pad:
                te_im[0:top_pad, left_pad:left_pad + c, :] = avg_chans
            if bottom_pad:
                te_im[r + top_pad:, left_pad:left_pad + c, :] = avg_chans
            if left_pad:
                te_im[:, 0:left_pad, :] = avg_chans
            if right_pad:
                te_im[:, c + left_pad:, :] = avg_chans
            im_patch = te_im[int(context_ymin):int(context_ymax + 1),
                       int(context_xmin):int(context_xmax + 1), :]
        else:
            im_patch = im[int(context_ymin):int(context_ymax + 1),
                       int(context_xmin):int(context_xmax + 1), :]

        if not np.array_equal(model_sz, original_sz):
            im_patch = cv2.resize(im_patch, (model_sz, model_sz))
        crop = im_patch.copy()
        im_patch = im_patch.transpose(2, 0, 1)
        im_patch = im_patch[np.newaxis, :, :, :]
        im_patch = im_patch.astype(np.float32)
        im_patch = torch.from_numpy(im_patch)
        im_patch = im_patch.cuda()
        return im_patch, crop


    def initialize_features(self,):
        #if not getattr(self, 'features_initialized', False):
            #self.net.initialize()
        self.features_initialized = True
        if self.net is None:
            from ltr.models.blockT.pvtv2_det_1 import build_pvtv2_det_pure_b3_tracker
            self.net = build_pvtv2_det_pure_b3_tracker(self.cross_dict)
            self.net.load_state_dict(torch.load(self.net_path)['net'], strict=False)
            self.net.cuda()


    def initialize(self, image, mask=None):
        self.frame_num =1
        tic = time.time()
        hanning = np.hanning(self.window_size)  # 14
        window = np.outer(hanning, hanning)
        self.window = window.flatten()
        self.window_penalty_func = self.window
        # Initialize
        self.initialize_features()
        """"""""""""
        if isinstance(mask, list):
            region = mask
        elif isinstance(mask, dict):
            bbox = mask['init_bbox']
            self.center_pos = np.array([bbox[0] + bbox[2] / 2,
                                        bbox[1] + bbox[3] / 2])
            self.size = np.array([bbox[2], bbox[3]])
            region = bbox

        else:
            region = rect_from_mask(mask)


        self.H, self.W, _ = image.shape
        gt_bbox_np = np.array(region).astype(np.float32)


        bbox = gt_bbox_np
        self.center_pos = np.array([bbox[0] + bbox[2] / 2,
                                    bbox[1] + bbox[3] / 2])
        self.size = np.array([bbox[2], bbox[3]])

        # calculate z crop size
        w_z = self.size[0] + (2 - 1) * ((self.size[0] + self.size[1]) * 0.5)  #
        h_z = self.size[1] + (2 - 1) * ((self.size[0] + self.size[1]) * 0.5)
        s_z = math.ceil(math.sqrt(w_z * h_z))

        # calculate channle average
        self.channel_average = np.mean(image, axis=(0, 1))

        # get crop
        z_crop, self.image_z = self.get_subwindow(image, self.center_pos,
                                    self.exemplar_size,
                                    s_z, self.channel_average)

        # normalize
        z_crop = z_crop.float().mul(1.0 / 255.0).clamp(0.0, 1.0)
        self.mean = [0.485, 0.456, 0.406]
        self.std = [0.229, 0.224, 0.225]
        self.inplace = False
        z_crop[0] = tvisf.normalize(z_crop[0], self.mean, self.std, self.inplace)

        # initialize template feature
        self.net.template(z_crop)
        out = {'time': time.time() - tic}
        self.z_c = z_crop


        return out

    def track(self, image, info=None, gt=None):
        self.video_name = info
        self.gt = gt
        self.bbox_prev = [self.center_pos[0] - self.size[0]*0.5, self.center_pos[1] - self.size[1]*0.5, self.size[0], self.size[1]]

        # calculate x crop size
        w_x = self.size[0] + (4 - 1) * ((self.size[0] + self.size[1]) * 0.5)
        h_x = self.size[1] + (4 - 1) * ((self.size[0] + self.size[1]) * 0.5)
        s_x = math.ceil(math.sqrt(w_x * h_x))

        # get crop
        x_crop, self.image_x = self.get_subwindow(image, self.center_pos,
                                    self.instance_size,
                                    round(s_x), self.channel_average)

        # normalize
        x_crop = x_crop.float().mul(1.0 / 255.0).clamp(0.0, 1.0)
        x_crop[0] = tvisf.normalize(x_crop[0], self.mean, self.std, self.inplace)

        # track
        outputs = self.net.track(x_crop)
        score = self._convert_score(outputs['pred_logits'])
        pred_bbox = self._convert_bbox(outputs['pred_boxes'])

        if int((score.size) ** 0.5) != self.window_size:
            self.window_size = int((score.size) ** 0.5)
            hanning = np.hanning(self.window_size)  # 14
            window = np.outer(hanning, hanning)
            self.window = window.flatten()
            self.window_penalty_func = self.window


        """ siamrpn post processing"""
        # score post-processing
        best_pscore_id, pscore, penalty = self._postprocess_score(
            score, pred_bbox, (self.bbox_prev[2], self.bbox_prev[3]), s_x)
        # box post-processing
        cx, cy, width, height = self._postprocess_box(
            best_pscore_id, score, pred_bbox, self.center_pos, self.size, s_x, penalty)
        """"""

        # window penalty
        pscore = score * (1 - self.window_penalty_func) + \
                 self.window * self.window_penalty_func

        best_idx = np.argmax(pscore)
        bbox = pred_bbox[:, best_idx]
        bbox_crop = bbox*256


        # clip boundary
        cx_clip, cy_clip, width_clip, height_clip = self._bbox_clip(cx, cy, width,
                                                height, image.shape[:2])

        # update state
        self.center_pos = np.array([cx_clip, cy_clip])
        self.size = np.array([width_clip, height_clip])

        bbox = [cx_clip - width_clip / 2,
                cy_clip - height_clip / 2,
                width_clip,
                height_clip]

        out = {'target_bbox': bbox,
               'best_score': pscore,
               'bbox_crop': bbox_crop}

        self.frame_num +=1

        """ save """
        import os
        import seaborn as sns
        import matplotlib.pyplot as plt
        import cv2
        cls_cp = score.reshape(32, 32)
        img = cv2.cvtColor(self.image_x, cv2.COLOR_RGB2BGR)
        full_img = cv2.cvtColor(image, cv2.COLOR_RGB2BGR)

        img_1 = img.copy()  # cv2.cvtColor(image, cv2.COLOR_RGB2BGR)
        bbox_crop = [bbox_crop[0] - 0.5 * bbox_crop[2], bbox_crop[1] - 0.5 * bbox_crop[3],
                     bbox_crop[0] + 0.5 * bbox_crop[2], bbox_crop[1] + 0.5 * bbox_crop[2]]
        full_predict = [bbox[0], bbox[1], bbox[0] + bbox[2], bbox[1] + bbox[3]]
        full_predict = list(map(int, full_predict))

        full_gt = [self.gt[0], self.gt[1], self.gt[0] + self.gt[2], self.gt[1] + self.gt[3]]
        full_gt = list(map(int, full_gt))

        pred_bbox = list(map(int, bbox_crop))
        cv2.rectangle(img, (pred_bbox[0], pred_bbox[1]),
                      (pred_bbox[2], pred_bbox[3]), (0, 255, 255), 2)
        cv2.putText(img, 'SBT', (20, 40), cv2.FONT_HERSHEY_SIMPLEX, 1, (0, 255, 255), 2)
        save_video_path = os.path.join('/home/v-xfei/Public/vis/video/', 'SBT', 'img', self.video_name)
        if not os.path.isdir(save_video_path):
            os.makedirs(save_video_path)
        result_img_path = os.path.join(save_video_path, '{}.jpg'.format(str(self.frame_num)))
        cv2.imwrite(result_img_path, img)

        cv2.rectangle(full_img, (full_gt[0], full_gt[1]),
                      (full_gt[2], full_gt[3]), (0, 0, 255), 2)
        cv2.rectangle(full_img, (full_predict[0], full_predict[1]),
                      (full_predict[2], full_predict[3]), (0, 255, 255), 2)
        cv2.putText(full_img, 'GT', (20, 40), cv2.FONT_HERSHEY_SIMPLEX, 1, (0, 0, 255), 2)
        cv2.putText(full_img, 'SBT', (80, 40), cv2.FONT_HERSHEY_SIMPLEX, 1, (0, 255, 255), 2)
        save_video_path = os.path.join('/home/v-xfei/Public/vis/video/', 'SBT', 'full_img', self.video_name)
        if not os.path.isdir(save_video_path):
            os.makedirs(save_video_path)
        result_img_path = os.path.join(save_video_path, '{}.jpg'.format(str(self.frame_num)))
        cv2.imwrite(result_img_path, full_img)

        sns.heatmap(cls_cp, vmin=0, vmax=1, cmap='viridis', cbar=False, annot=False, xticklabels=False,
                    yticklabels=False)  # YlGnBu_r
        save_video_path = os.path.join('/home/v-xfei/Public/vis/video/', 'SBT', 'heatmap', self.video_name)
        if not os.path.isdir(save_video_path):
            os.makedirs(save_video_path)
        # plt.show()
        plt.savefig(os.path.join(save_video_path, str(self.frame_num) + '.png'))
        plt.close()

        cls_pad = cls_cp  # np.zeros([37, 37])

        # cls_pad[9:-9, 9:-9] = cls_cp

        def vis_heatmap(img, att):
            import cv2
            import numpy as np

            """ 
            img: np.ndarray (uint8) in shape of [H, W, 3]. 
            att: np.ndarray (float32) in shape of [H, W], ranged [0, 1]
            """
            h, w = img.shape[0:2]
            att = cv2.resize(att, (w, h), interpolation=cv2.INTER_CUBIC)
            att = np.clip(att, 0, 1)

            att_uint8 = (att * 255).astype(np.uint8)
            att_vis = cv2.applyColorMap(att_uint8, cv2.COLORMAP_JET)

            img = img.astype(np.float32)
            att_vis = att_vis.astype(np.float32)

            out = (1 - att ** 0.8).reshape(h, w, 1) * img + \
                  (att ** 0.8).reshape(h, w, 1) * att_vis

            out = np.clip(out, 0, 255).astype(np.uint8)
            return out

        img_cls = vis_heatmap(img_1, cls_pad)
        save_video_path = os.path.join('/home/v-xfei/Public/vis/video/', 'SBT', 'img_heatmap', self.video_name)
        if not os.path.isdir(save_video_path):
            os.makedirs(save_video_path)
        cv2.rectangle(img_cls, (pred_bbox[0], pred_bbox[1]),
                      (pred_bbox[2], pred_bbox[3]), (0, 255, 255), 2)
        cv2.putText(img_cls, 'SBT', (20, 40), cv2.FONT_HERSHEY_SIMPLEX, 1, (0, 255, 255), 2)
        result_img_path = os.path.join(save_video_path, '{}.jpg'.format(str(self.frame_num)))
        cv2.imwrite(result_img_path, img_cls)

        """ save end """

        return out

    def _bbox_clip_xyxy(self, x1, y1, x2, y2, boundary, min_sz=10):
        '''boundary (H,W)'''
        x1_new = max(0, min(x1, boundary[1] - min_sz))
        y1_new = max(0, min(y1, boundary[0] - min_sz))
        x2_new = max(min_sz, min(x2, boundary[1]))
        y2_new = max(min_sz, min(y2, boundary[0]))
        return x1_new, y1_new, x2_new, y2_new

    def _postprocess_score(self, score, box_wh, target_sz, scale_x):
        r"""
        Perform SiameseRPN-based tracker's post-processing of score
        :param score: (HW, ), score prediction
        :param box_wh: (HW, 4), cxywh, bbox prediction (format changed)
        :param target_sz: previous state (w & h)
        :param scale_x:
        :return:
            best_pscore_id: index of chosen candidate along axis HW
            pscore: (HW, ), penalized score
            penalty: (HW, ), penalty due to scale/ratio change
        """
        def change(r):
            return np.maximum(r, 1. / r)

        def sz(w, h):
            pad = (w + h) * 0.5
            sz2 = (w + pad) * (h + pad)
            return np.sqrt(sz2)

        def sz_wh(wh):
            pad = (wh[0] + wh[1]) * 0.5
            sz2 = (wh[0] + pad) * (wh[1] + pad)
            return np.sqrt(sz2)

        # size penalty
        penalty_k = self.penalty_k
        target_sz_in_crop = [target_sz[0] , target_sz[1] ]
        s_c = change(
            sz(box_wh[2, :]*scale_x, box_wh[3, :]*scale_x) /
            (sz_wh(target_sz_in_crop)))  # scale penalty
        r_c = change((target_sz_in_crop[0] / target_sz_in_crop[1]) /
                     (box_wh[2, :] / box_wh[3, :]))  # ratio penalty
        penalty = np.exp(-(r_c * s_c - 1) * penalty_k)
        pscore = penalty * score

        # ipdb.set_trace()
        # cos window (motion model)
        window_influence = self.window_influence
        pscore = pscore * (
            1 - window_influence) + self.window_penalty_func * window_influence
        best_pscore_id = np.argmax(pscore)

        return best_pscore_id, pscore, penalty

    def _postprocess_box(self, best_pscore_id, score, box_wh, target_pos,
                         target_sz, scale_x, penalty):
        r"""
        Perform SiameseRPN-based tracker's post-processing of box
        :param score: (HW, ), score prediction
        :param box_wh: (HW, 4), cxywh, bbox prediction (format changed)
        :param target_pos: (2, ) previous position (x & y)
        :param target_sz: (2, ) previous state (w & h)
        :param scale_x: scale of cropped patch of current frame
        :param x_size: size of cropped patch
        :param penalty: scale/ratio change penalty calculated during score post-processing
        :return:
            new_target_pos: (2, ), new target position
            new_target_sz: (2, ), new target size
        """
        pred_in_crop = box_wh[:, best_pscore_id] * np.float32(scale_x)
        # about np.float32(scale_x)
        # attention!, this casting is done implicitly
        # which can influence final EAO heavily given a model & a set of hyper-parameters

        # box post-postprocessing
        test_lr = self.test_lr
        lr = penalty[best_pscore_id] * score[best_pscore_id] * test_lr
        res_x = pred_in_crop[0] + target_pos[0] - scale_x / 2 #pred_in_crop[0] + target_pos[0] - (x_size // 2) / scale_x
        res_y = pred_in_crop[1] + target_pos[1] - scale_x / 2 #pred_in_crop[1] + target_pos[1] - (x_size // 2) / scale_x
        res_w = target_sz[0] * (1 - lr) + pred_in_crop[2] * lr
        res_h = target_sz[1] * (1 - lr) + pred_in_crop[3] * lr

        return res_x, res_y, res_w, res_h

    def _restrict_box(self, target_pos, target_sz):
        r"""
        Restrict target position & size
        :param target_pos: (2, ), target position
        :param target_sz: (2, ), target size
        :return:
            target_pos, target_sz
        """
        target_pos[0] = max(0, min(self._state['im_w'], target_pos[0]))
        target_pos[1] = max(0, min(self._state['im_h'], target_pos[1]))
        target_sz[0] = max(self._hyper_params['min_w'],
                           min(self._state['im_w'], target_sz[0]))
        target_sz[1] = max(self._hyper_params['min_h'],
                           min(self._state['im_h'], target_sz[1]))

        return target_pos, target_sz

    def _cvt_box_crop2frame(self, box_in_crop, target_pos, scale_x, x_size):
        r"""
        Convert box from cropped patch to original frame
        :param box_in_crop: (4, ), cxywh, box in cropped patch
        :param target_pos: target position
        :param scale_x: scale of cropped patch
        :param x_size: size of cropped patch
        :return:
            box_in_frame: (4, ), cxywh, box in original frame
        """
        x = (box_in_crop[..., 0]) / scale_x + target_pos[0] - (x_size //
                                                               2) / scale_x
        y = (box_in_crop[..., 1]) / scale_x + target_pos[1] - (x_size //
                                                               2) / scale_x
        w = box_in_crop[..., 2] / scale_x
        h = box_in_crop[..., 3] / scale_x
        box_in_frame = np.stack([x, y, w, h], axis=-1)

        return box_in_frame












def run_vot_exp(VIS=False):

    torch.set_num_threads(1)
    # torch.cuda.set_device(CUDA_ID)  # set GPU id
    save_root = '/home/v-xfei/Public/dataset/vot_tools/vot21/vis/'
    net_path = '/home/v-xfei/Public/project/TransT_result/alldata/pvt_v2_b3_ep0190.pth.tar'
    if VIS and (not os.path.exists(save_root)):
        os.mkdir(save_root)
    crossdict = [
        [2],
        [2, 4],
        [2, 4, 6, 8, 10],
    ]
    # model = build_pvtv2b3_tracker(crossdict)
    from ltr.models.blockT1.pvtv2b3 import build_pvtv2b3_tracker
    model = build_pvtv2b3_tracker(crossdict)

    model.load_state_dict(torch.load(net_path)['net'], strict=False)
    model.cuda()

    tracker = Tracker_blockT(name='Tracker_blockT', net_path=net_path, type=args.type, window_penalty=0.49,
                               window_size=32, exemplar_size=128, instance_size=256, net=model)
    handle = vot.VOT("mask")
    selection = handle.region()
    imagefile = handle.frame()

    if VIS:
        '''for vis'''
        seq_name = imagefile.split('/')[-3]
        save_v_dir = os.path.join(save_root,seq_name)
        if not os.path.exists(save_v_dir):
            os.mkdir(save_v_dir)
        cur_time = int(time.time() % 10000)
        save_dir = os.path.join(save_v_dir, str(cur_time))
        if not os.path.exists(save_dir):
            os.mkdir(save_dir)

    image = cv2.cvtColor(cv2.imread(imagefile), cv2.COLOR_BGR2RGB) # Right
    # mask given by the toolkit ends with the target (zero-padding to the right and down is needed)
    mask = make_full_size(selection, (image.shape[1], image.shape[0]))
    tracker.initialize(image, mask)

    while True:
        imagefile = handle.frame()
        if not imagefile:
            break
        image = cv2.cvtColor(cv2.imread(imagefile), cv2.COLOR_BGR2RGB)  # Right
        b1, m = tracker.track(image)
        handle.report(m)
        if VIS:
            '''Visualization'''
            image_show = image[:,:,::-1].copy() # RGB --> BGR
            image_show = image_show.astype(np.float32)

            image_show[:,:,0] += 127.0 * m
            image_show = image_show.clip(0,255).astype(np.uint8)
            cv2.rectangle(image_show,(int(b1[0]),int(b1[1])),
                          (int(b1[0]+b1[2]),int(b1[1]+b1[3])),(0,0,255),2)
            save_path = os.path.join(save_dir, imagefile.split('/')[-1])
            cv2.imwrite(save_path,image_show)

#run_vot_exp()



        """ save """
        import os
        import seaborn as sns
        import matplotlib.pyplot as plt
        import cv2
        cls_cp = score.reshape(32, 32)
        img = cv2.cvtColor(self.image_x, cv2.COLOR_RGB2BGR)
        full_img = cv2.cvtColor(image, cv2.COLOR_RGB2BGR)

        img_1 = img.copy()  # cv2.cvtColor(image, cv2.COLOR_RGB2BGR)
        bbox_crop = [bbox_crop[0]-0.5*bbox_crop[2], bbox_crop[1]-0.5*bbox_crop[3], bbox_crop[0]+0.5*bbox_crop[2], bbox_crop[1]+0.5*bbox_crop[2]]
        full_predict = [bbox[0], bbox[1], bbox[0] + bbox[2], bbox[1] + bbox[3]]
        full_predict = list(map(int, full_predict))

        full_gt = [self.gt[0], self.gt[1], self.gt[0] + self.gt[2], self.gt[1] + self.gt[3]]
        full_gt = list(map(int, full_gt))

        pred_bbox = list(map(int, bbox_crop))
        cv2.rectangle(img, (pred_bbox[0], pred_bbox[1]),
                      (pred_bbox[2], pred_bbox[3]), (0, 255, 255), 2)
        cv2.putText(img, 'SBT', (20, 40), cv2.FONT_HERSHEY_SIMPLEX, 1, (0, 0, 255), 2)
        save_video_path = os.path.join('/home/v-xfei/Public/vis/video/', 'sbt', 'img', self.video_name)
        if not os.path.isdir(save_video_path):
            os.makedirs(save_video_path)
        result_img_path = os.path.join(save_video_path, '{}.jpg'.format(str(self.frame_num)))
        cv2.imwrite(result_img_path, img)


        cv2.rectangle(full_img, (full_gt[0], full_gt[1]),
                      (full_gt[2], full_gt[3]), (0, 0, 255), 2)
        cv2.rectangle(full_img, (full_predict[0], full_predict[1]),
                      (full_predict[2], full_predict[3]), (0, 255, 255), 2)
        cv2.putText(full_img, 'GT', (20, 40), cv2.FONT_HERSHEY_SIMPLEX, 1, (0, 0, 255), 2)
        cv2.putText(full_img, 'SBT', (80, 40), cv2.FONT_HERSHEY_SIMPLEX, 1, (0, 255, 255), 2)
        save_video_path = os.path.join('/home/v-xfei/Public/vis/video/', 'sbt', 'full_img', self.video_name)
        if not os.path.isdir(save_video_path):
            os.makedirs(save_video_path)
        result_img_path = os.path.join(save_video_path, '{}.jpg'.format(str(self.frame_num)))
        cv2.imwrite(result_img_path, full_img)


        sns.heatmap(cls_cp, vmin=0, vmax=1, cmap='viridis', cbar=False, annot=False, xticklabels=False,
                    yticklabels=False)  # YlGnBu_r
        save_video_path = os.path.join('/home/v-xfei/Public/vis/video/', 'sbt', 'heatmap', self.video_name)
        if not os.path.isdir(save_video_path):
            os.makedirs(save_video_path)
        # plt.show()
        plt.savefig(os.path.join(save_video_path, str(self.frame_num) + '.png'))
        plt.close()

        cls_pad = cls_cp# np.zeros([37, 37])
        #cls_pad[9:-9, 9:-9] = cls_cp

        def vis_heatmap(img, att):
            import cv2
            import numpy as np

            """ 
            img: np.ndarray (uint8) in shape of [H, W, 3]. 
            att: np.ndarray (float32) in shape of [H, W], ranged [0, 1]
            """
            h, w = img.shape[0:2]
            att = cv2.resize(att, (w, h), interpolation=cv2.INTER_CUBIC)
            att = np.clip(att, 0, 1)

            att_uint8 = (att * 255).astype(np.uint8)
            att_vis = cv2.applyColorMap(att_uint8, cv2.COLORMAP_JET)

            img = img.astype(np.float32)
            att_vis = att_vis.astype(np.float32)

            out = (1 - att ** 0.8).reshape(h, w, 1) * img + \
                  (att ** 0.8).reshape(h, w, 1) * att_vis

            out = np.clip(out, 0, 255).astype(np.uint8)
            return out

        img_cls = vis_heatmap(img_1, cls_pad)
        save_video_path = os.path.join('/home/v-xfei/Public/vis/video/', 'sbt', 'img_heatmap', self.video_name)
        if not os.path.isdir(save_video_path):
            os.makedirs(save_video_path)
        cv2.rectangle(img_cls, (pred_bbox[0], pred_bbox[1]),
                      (pred_bbox[2], pred_bbox[3]), (0, 255, 255), 2)
        cv2.putText(img_cls, 'SBT', (20, 40), cv2.FONT_HERSHEY_SIMPLEX, 1, (0, 255, 255), 2)
        result_img_path = os.path.join(save_video_path, '{}.jpg'.format(str(self.frame_num)))
        cv2.imwrite(result_img_path, img_cls)

        """ save end """