from __future__ import absolute_import

import numpy as np
import math
import torchvision.transforms.functional as tvisf
import cv2
import torch
import torch.nn.functional as F
import time
import random
import os
import trax

from pysot_toolkit import vot2020 as vot
from pytracking.external.AR_VOT21.pytracking.ARcm_seg import ARcm_seg

from pysot_toolkit.trackers.vot20_utils import *


class Tracker_SBTv1(object):

    def __init__(self, name, net_path=None,type='base', window_penalty=0.49, window_size=16, exemplar_size=128,
                 instance_size=256, net =None, refine_path=None
                 ):
        self.name = name
        self.net_path = net_path
        self.exemplar_size = exemplar_size
        self.instance_size = instance_size
        self.net = None
        self.window_size = int(window_size)
        self.type= type

        self.THRES = 0.65
        self.window_penalty = window_penalty

        self.penalty_k = 0.04
        self.test_lr = 0.95
        self.window_influence = 0.21
        self.net = net

        self.refine_path = refine_path
        if self.refine_path:
            self.alpha = ARcm_seg(refine_path, input_sz=384)

    def _convert_score(self, score):

        score = score.permute(2, 1, 0).contiguous().view(2, -1).permute(1, 0)
        score = F.softmax(score, dim=1).data[:, 0].cpu().numpy()
        return score

    def _convert_bbox(self, delta):

        delta = delta.permute(2, 1, 0).contiguous().view(4, -1)
        delta = delta.data.cpu().numpy()

        return delta

    def _bbox_clip(self, cx, cy, width, height, boundary):
        cx = max(0, min(cx, boundary[1]))
        cy = max(0, min(cy, boundary[0]))
        width = max(10, min(width, boundary[1]))
        height = max(10, min(height, boundary[0]))
        return cx, cy, width, height

    def _bbox_clip_xyxy(self, x1, y1, x2, y2, boundary, min_sz=10):
        '''boundary (H,W)'''
        x1_new = max(0, min(x1, boundary[1] - min_sz))
        y1_new = max(0, min(y1, boundary[0] - min_sz))
        x2_new = max(min_sz, min(x2, boundary[1]))
        y2_new = max(min_sz, min(y2, boundary[0]))
        return x1_new, y1_new, x2_new, y2_new

    def get_subwindow(self, im, pos, model_sz, original_sz, avg_chans):
        """
        args:
            im: rgb based image
            pos: center position
            model_sz: exemplar size
            original_sz: original size
            avg_chans: channel average
        """
        if isinstance(pos, float):
            pos = [pos, pos]
        sz = original_sz
        im_sz = im.shape
        c = (original_sz + 1) / 2
        # context_xmin = round(pos[0] - c) # py2 and py3 round
        context_xmin = np.floor(pos[0] - c + 0.5)
        context_xmax = context_xmin + sz - 1
        # context_ymin = round(pos[1] - c)
        context_ymin = np.floor(pos[1] - c + 0.5)
        context_ymax = context_ymin + sz - 1
        left_pad = int(max(0., -context_xmin))
        top_pad = int(max(0., -context_ymin))
        right_pad = int(max(0., context_xmax - im_sz[1] + 1))
        bottom_pad = int(max(0., context_ymax - im_sz[0] + 1))

        context_xmin = context_xmin + left_pad
        context_xmax = context_xmax + left_pad
        context_ymin = context_ymin + top_pad
        context_ymax = context_ymax + top_pad

        r, c, k = im.shape
        if any([top_pad, bottom_pad, left_pad, right_pad]):
            size = (r + top_pad + bottom_pad, c + left_pad + right_pad, k)
            te_im = np.zeros(size, np.uint8)
            te_im[top_pad:top_pad + r, left_pad:left_pad + c, :] = im
            if top_pad:
                te_im[0:top_pad, left_pad:left_pad + c, :] = avg_chans
            if bottom_pad:
                te_im[r + top_pad:, left_pad:left_pad + c, :] = avg_chans
            if left_pad:
                te_im[:, 0:left_pad, :] = avg_chans
            if right_pad:
                te_im[:, c + left_pad:, :] = avg_chans
            im_patch = te_im[int(context_ymin):int(context_ymax + 1),
                       int(context_xmin):int(context_xmax + 1), :]
        else:
            im_patch = im[int(context_ymin):int(context_ymax + 1),
                       int(context_xmin):int(context_xmax + 1), :]

        if not np.array_equal(model_sz, original_sz):
            im_patch = cv2.resize(im_patch, (model_sz, model_sz))
        crop = im_patch.copy()
        im_patch = im_patch.transpose(2, 0, 1)
        im_patch = im_patch[np.newaxis, :, :, :]
        im_patch = im_patch.astype(np.float32)
        im_patch = torch.from_numpy(im_patch)
        im_patch = im_patch.cuda()
        return im_patch, crop


    def initialize(self, image, bbox_input=None, input_type='4bbox'):  # 8bbox, 4bbox, mask
        self.frame_num =1
        tic = time.time()
        hanning = np.hanning(self.window_size)  # 14
        window = np.outer(hanning, hanning)
        self.window = window.flatten()
        self.window_penalty_func = self.window
        # Initialize
        """"""""""""
        if input_type == '4bbox':
            self.center_pos = np.array([bbox_input[0] + bbox_input[2] / 2, bbox_input[1] + bbox_input[3] / 2])
            self.size = np.array([bbox_input[2], bbox_input[3]])
            region = bbox_input

        elif input_type=='mask':
            region = rect_from_mask(bbox_input)
            self.first_mask = bbox_input[region[1]: region[1] + region[3], region[0]:region[0] + region[2]]

        elif input_type == '8bbox':
            pass


        self.H, self.W, _ = image.shape
        gt_bbox_np = np.array(region).astype(np.float32)

        if self.refine_path:
            '''initilize refinement module for specific video'''
            self.alpha.initialize(image, gt_bbox_np)

        bbox = gt_bbox_np
        self.center_pos = np.array([bbox[0] + bbox[2] / 2,
                                    bbox[1] + bbox[3] / 2])
        self.size = np.array([bbox[2], bbox[3]])

        # calculate z crop size
        w_z = self.size[0] + (2 - 1) * ((self.size[0] + self.size[1]) * 0.5)  #
        h_z = self.size[1] + (2 - 1) * ((self.size[0] + self.size[1]) * 0.5)
        s_z = math.ceil(math.sqrt(w_z * h_z))

        # calculate channle average
        self.channel_average = np.mean(image, axis=(0, 1))

        # get crop
        z_crop, self.image_z = self.get_subwindow(image, self.center_pos,
                                    self.exemplar_size,
                                    s_z, self.channel_average)

        # normalize
        z_crop = z_crop.float().mul(1.0 / 255.0).clamp(0.0, 1.0)
        self.mean = [0.485, 0.456, 0.406]
        self.std = [0.229, 0.224, 0.225]
        self.inplace = False
        z_crop[0] = tvisf.normalize(z_crop[0], self.mean, self.std, self.inplace)

        # initialize template feature
        self.net.template(z_crop)
        out = {'time': time.time() - tic}
        return out

    def track(self, image, info=None):
        self.bbox_prev = [self.center_pos[0] - self.size[0]*0.5, self.center_pos[1] - self.size[1]*0.5, self.size[0], self.size[1]]

        # calculate x crop size
        w_x = self.size[0] + (4 - 1) * ((self.size[0] + self.size[1]) * 0.5)
        h_x = self.size[1] + (4 - 1) * ((self.size[0] + self.size[1]) * 0.5)
        s_x = math.ceil(math.sqrt(w_x * h_x))

        # get crop
        x_crop, self.image_x = self.get_subwindow(image, self.center_pos,
                                    self.instance_size,
                                    round(s_x), self.channel_average)

        # normalize
        x_crop = x_crop.float().mul(1.0 / 255.0).clamp(0.0, 1.0)
        x_crop[0] = tvisf.normalize(x_crop[0], self.mean, self.std, self.inplace)

        # track
        outputs = self.net.track(x_crop)
        score = self._convert_score(outputs['pred_logits'])
        pred_bbox = self._convert_bbox(outputs['pred_boxes'])

        if int((score.size) ** 0.5) != self.window_size:
            self.window_size = int((score.size) ** 0.5)
            hanning = np.hanning(self.window_size)  # 14
            window = np.outer(hanning, hanning)
            self.window = window.flatten()
            self.window_penalty_func = self.window

        # window penalty
        pscore = score * (1 - self.window_penalty) + self.window * self.window_penalty

        best_idx = np.argmax(pscore)
        bbox = pred_bbox[:, best_idx]
        bbox = bbox * s_x
        cx = bbox[0] + self.center_pos[0] - s_x / 2
        cy = bbox[1] + self.center_pos[1] - s_x / 2
        width = bbox[2]
        height = bbox[3]

        # clip boundary
        cx, cy, width, height = self._bbox_clip(cx, cy, width, height, image.shape[:2])

        # update state
        self.center_pos = np.array([cx, cy])
        self.size = np.array([width, height])

        bbox = [cx - width / 2,
                cy - height / 2,
                width,
                height]

        out = {'target_bbox': bbox,
                   'best_score': pscore,
               }
        self.frame_num +=1

        if self.refine_path:
            '''Mask'''
            pred_mask, search, search_mask = self.alpha.get_mask(image, np.array(bbox), vis=F, trt='false')
            #pred_mask = self.alpha.refine(image, np.array(bbox))
            final_mask = (pred_mask > self.THRES).astype(np.uint8)

            out ['mask'] = final_mask

        return out



def set_seed(seed):
    random.seed(seed)
    os.environ['PYTHONHASHSEED'] = str(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.backends.cudnn.benchmark = True
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.enabled = True


def run_vot_exp_box():

    seed_test = 1000
    set_seed(seed_test)
    torch.set_num_threads(1)
    # torch.cuda.set_device(CUDA_ID)  # set GPU id
    from ltr.models.SBT.SBTv1 import build_SBTv1_base_model
    net_path = '/home/jaffe/Titanspace/SBTv2/vot_pth/checkpoints/ltr/SBTvot/SBTv1_jaf/SBTv1_base_ep0540.pth.tar'
    refine_path = None

    net = build_SBTv1_base_model()
    net.load_state_dict(torch.load(net_path)['net'])
    net.cuda()
    tracker_nc = Tracker_SBTv1(name='SBTv1_520', net=net, window_penalty=0.49, window_size=32, exemplar_size=128, instance_size=256, refine_path=refine_path)
    tracker = tracker_nc

    handle = vot.VOT(trax.Region.RECTANGLE)
    vot_anno = handle.region()
    imagefile = handle.frame()

    def _convert_anno_to_list(vot_anno):
        vot_anno = [vot_anno[0], vot_anno[1], vot_anno[2], vot_anno[3]]
        return vot_anno

    init_box = _convert_anno_to_list(vot_anno)
    image = cv2.cvtColor(cv2.imread(imagefile), cv2.COLOR_BGR2RGB) # Right
    tracker.initialize(image, init_box)

    while True:
        imagefile = handle.frame()
        if not imagefile:
            break
        image = cv2.cvtColor(cv2.imread(imagefile), cv2.COLOR_BGR2RGB)  # Right
        out = tracker.track(image)
        bbox = out['target_bbox']
        handle.report(vot.Rectangle(bbox[0], bbox[1], bbox[2], bbox[3]))


if __name__ == '__main__':
    run_vot_exp_box()